import math

import torch
from torch import nn
import torch.nn.functional as F

class ExpRNN(nn.Module):
    def __init__(self, input_size, hidden_size, **kw):
	    raise NotImplementedError("see `exprnn` branch")

def rnn_shim(cls):
    """LSTM API for GRU and RNN.
    
    hidden state is first element of state tuple"""
    class shim(cls):
        def forward(self, input, states=(None,)):
            assert len(states)==1
            out, h = super().forward(input, *states)
            return out, (h,)
    return shim

GRU = rnn_shim(nn.GRU)
RNN = rnn_shim(nn.RNN)
LSTM = nn.LSTM


class GenericRNN(nn.Module):
    kind_cls = {
        'gru':GRU,
        'lstm':LSTM,
        'elman':RNN,
        'exprnn':ExpRNN
        }
    def __init__(self, kind, *a, **kw):
        super().__init__()
        if kw.get('bidirectional'): raise ValueError("""
            bidirectional GenericRNN not supported.
            """)
        cls = GenericRNN.kind_cls[kind]
        self.kind = kind
        self.rnn = cls(*a, **kw)

    def __getattr__(self, a):
        try:
            return  super().__getattr__(a)
        except AttributeError:
            return getattr(self.rnn, a)

    def forward(self, x, initial_state):
        """
        Args:
            x: Tensor[batch x time x channel] if batch_first else [time x batch x channel]
            initial_state: List[Tensor[layers x batch x hidden]]], list of components 
            with 0 being hidden state (e.g. 1 is cell state for LSTM). 
        Returns:
            hidden: hidden states of top layers Tensor[batch x time x hidden]
                or [time x batch x hidden]
            new_states: List[Tensor[layers x batch x hidden]]
        """
        hidden, final_state = self.rnn.forward(x, initial_state)  #forward or __call__?
        return hidden, final_state

    ## NOTE: individual time-step API might be useful, not actually needed yet though
    # def step(self, x, state):
    #     """
    #     Args:
    #         x: Tensor[batch x channel]
    #         state: List[Tensor[layers x batch x hidden]]], list of components 
    #         with 0 being hidden state (e.g. 1 is cell state for LSTM). 
    #     Returns:
    #         hidden: hidden state of top layer [batch x hidden]
    #         new_states: List[Tensor[layers x batch x hidden]]
    #     """
    #     time_idx = 1 if self.rnn.batch_first else 0
    #     x = x.unsqueeze(time_idx)
    #     hidden, state = self.forward(x, state)
    #     return hidden.squeeze(time_idx), state


import math
import torch
from torch import nn
import torch.distributions as D
import torch.nn.functional as F

def reweight_quantile(probs, min_q=0, max_q=1):
    """
    reweight ordinal discrete distribution to have mass only between quantiles
    """
    # TODO
    cdf = probs.cumsum(-1)
    raise NotImplementedError


def reweight_top_p(probs, top_p):
    """
    given tensor of probabilities, apply top p / "nucleus" filtering,
    or temperature if `top_p` is greater than 1
    """
    if top_p > 1:
        probs = probs**(1/top_p)
        return probs / probs.sum(-1)

    # NOTE: this is fudged slightly, it doesn't 'interpolate' the cutoff bin
    desc_probs, idx = probs.sort(-1, descending=True)
    iidx = idx.argsort(-1)
    cumprob = desc_probs.cumsum(-1)
    # first index where cumprob >= top_p is the last index we don't zero
    to_zero = (cumprob >= top_p).roll(1, -1)
    to_zero[...,0] = False
    # unsort
    to_zero = to_zero.gather(-1, iidx)
    weighted_probs = torch.zeros_like(probs).where(to_zero, probs)
    return weighted_probs / weighted_probs.sum(-1, keepdim=True)


def categorical_sample(
        logits, whitelist=None, index=None, top_p=None, 
        truncate_quantile=None
        ):
    """if whitelist is a dictionary, it maps to weights"""
    # if logits.isnan().any():
    #     raise Exception('start '+str(logits))

    if whitelist is not None:
        idx = list(whitelist)
        preserve_logits = logits[...,idx]
        if isinstance(whitelist, dict):
            preserve_logits += torch.tensor(list(whitelist.values())).log()
        logits = torch.full_like(logits, -torch.inf)
        logits[..., idx] = preserve_logits

    if index is not None:
        return logits.argsort(-1, True)[..., index]
    
    probs = logits.softmax(-1)

    # if probs.isnan().any():
        # raise Exception('whitelist '+str(probs))
    
    if top_p is not None:
        probs = reweight_top_p(probs, top_p)

    # if probs.isnan().any():
        # raise Exception('top p '+str(probs))

    if truncate_quantile is not None:
        q_lo, q_hi = truncate_quantile
        q_lo = max(0., 0. if q_lo is None else q_lo)
        q_hi = min(1., 1. if q_hi is None else q_hi)
        # print(q_lo, q_hi)
        zero = torch.zeros_like(probs)
        zcs = (1-probs.flip(-1).cumsum(-1).flip(-1))
        cs = probs.cumsum(-1)
        # truncate up to q_lo
        probs -= (q_lo-zcs).clip(zero, probs)
        # truncate from q_hi
        probs -= (cs-q_hi).clip(zero, probs)

    # if probs.isnan().any():
        # raise Exception('trunc quant '+str(probs))

    probs.nan_to_num_(1e-5)

    return D.Categorical(probs).sample()


class CensoredMixtureLogistic(nn.Module):
    def __init__(self, n, res=1e-2, lo='-inf', hi='inf', 
            sharp_bounds=(1e-4,2e3), init=None):
        super().__init__()
        self.n = n
        self.res = res
        self.sharp_bounds = sharp_bounds
        self.register_buffer('lo', torch.tensor(float(lo)))
        self.register_buffer('hi', torch.tensor(float(hi)))
        # TODO: init is not general-purpose
        if init=='time':
            self.bias = nn.Parameter(torch.cat((
                torch.zeros(n), torch.logspace(-3,1,n), torch.zeros(n)
                )))
        elif init=='velocity':
            self.bias = nn.Parameter(torch.cat((
                torch.zeros(n), torch.linspace(0,127,n), torch.zeros(n)
                )))
        else:
            self.bias = nn.Parameter(torch.cat((
                torch.zeros(n), torch.randn(n), torch.zeros(n)
                )))

    @property
    def n_params(self):
        return self.n*3

    def get_params(self, h):
        assert h.shape[-1] == self.n_params
        h = h+self.bias
        # get parameters from unconstrained hidden state:
        logit_pi, loc, log_s = torch.chunk(h, 3, -1)
        # mixture coefficients
        log_pi = logit_pi - logit_pi.logsumexp(-1,keepdim=True)
        # location
        loc = loc.clamp(self.lo-10*self.res, self.hi+10*self.res)
        # sharpness
        s = F.softplus(log_s).clamp(*self.sharp_bounds)
        return log_pi, loc, s


    def forward(self, h, x):
        """log prob of x under distribution parameterized by h
        Args:
            h: Tensor[...,n_params]
            x: Tensor[...]
            "..." dims must broadcast
        """
        log_pi, loc, s = self.get_params(h)    

        d = self.res/2
        x = x.clamp(self.lo, self.hi)[...,None]
        x_ = (x - loc) * s
        sd = s*d

        # # censoring
        lo_cens = x <= self.lo+d
        hi_cens = x >= self.hi-d
        ones = torch.ones_like(s)
        zeros = torch.zeros_like(s)

        diff_term = torch.where(lo_cens | hi_cens, 
            ones, sd.exp() - (-sd).exp()
            ).log()
        minus_sp_term = torch.where(hi_cens, -sd, F.softplus(-sd-x_))
        plus_sp_term = torch.where(lo_cens, zeros, x_ + F.softplus(sd-x_))

        log_delta_cdf = diff_term - minus_sp_term - plus_sp_term
    
        # log prob
        r = {
            'log_prob': (log_pi + log_delta_cdf).logsumexp(-1)
        }
        # diagnostics
        with torch.no_grad():
            ent = D.Categorical(logits=log_pi).entropy()
            r |= {
                # 'min_sharpness': s.min(),
                'max_sharpness': s.max(),
                'mean_sharpness': (s*log_pi.exp()).sum(-1).mean(),
                # 'min_entropy': ent.min(),
                # 'max_entropy': ent.max(),
                'mean_cmp_entropy': ent.mean(),
                'marginal_cmp_entropy': D.Categorical(
                    log_pi.exp().mean(list(range(log_pi.ndim-1)))).entropy(),
                # 'min_loc': loc.min(),
                # 'max_loc': loc.max()
            }
        return r

    def cdf(self, h, x):
        """
        Args:
            h: Tensor[...,n_params]
            x: Tensor[...]
                `h` should broadcast with `x[...,None]`
        Returns:
            cdf: Tensor[...] (shape of `x` broadcasted with `h[...,0]`)
        """
        if not isinstance(x, torch.Tensor):
            x = torch.tensor(x)
        log_pi, loc, s = self.get_params(h)  
        cdfs = self.cdf_components(loc, s, x)
        cdf = (cdfs * log_pi.softmax(-1)).sum(-1)
        return cdf

    def cdf_components(self, loc, s, x):
        x_ = (x[...,None] - loc) * s
        return x_.sigmoid()       

    # TODO: 'discrete_sample' method which would re-quantize and then allow
    # e.g. nucleus sampling on the categorical distribution?
    def sample(self, h, truncate=None, shape=None, 
        weight_top_p=None, component_temp=None, bias=None, 
        truncate_quantile=None, quantile_k=128, eps=1e-9
        ):
        """
        Args:
            h: Tensor[..., n_params]
            truncate: Optional[Tuple[float, float]]. lower and upper bound for truncation.
            shape: Optional[int]. additional sample shape to be prepended to dims.
            weight_top_p: top_p ("nucleus") filtering for mixture weights.
                default is 1 (no change to distribution). 0 would sample top
                component (after truncation) only.
            component_temp: Optional[float]. sampling temperature of each mixture 
                component. default is 1. 0 would sample component location only,
                ignoring sharpness.
            bias: applied outside of truncation but inside of clamping,
                useful e.g. for latency correction when sampling delta-time
            truncate_quantile: truncate the distribution again by quantile,
                after the effects of truncate, weight_top_p and component_temp.
            quantile_k: truncate_quantile is implemented by drawing this many
                samples and sorting them  
        Returns:
            Tensor[*shape,...] (h without last dimension, prepended with `shape`)
        """
        if truncate_quantile == (0,1):
            truncate_quantile = None

        if truncate is None:
            truncate = (-torch.inf, torch.inf)
        # early out in the single possibility case
        if truncate[0] == truncate[1]:
            return torch.tensor(
                [truncate[0]]*shape if shape is not None else truncate[0])
        truncate = torch.tensor(truncate)

        if shape is None:
            unwrap = True
            shape = 1
        else:
            unwrap = False

        if truncate_quantile is not None:
            # draw k samples
            shape = shape * quantile_k

        if component_temp is None:
            component_temp = 1

        if bias is None:
            bias = 0

        log_pi, loc, s = self.get_params(h)
        s = s/component_temp
        scale = 1/s

        # cdfs: [...,bound,component]
        cdfs = self.cdf_components(loc[...,None,:], s[...,None,:], truncate) 
        # prob. mass of each component within bounds
        trunc_probs = cdfs[...,1,:] - cdfs[...,0,:] # [...,component]
        probs = log_pi.exp() * trunc_probs # reweighted mixture component probs
        if weight_top_p is not None:
            # reweight with top_p
            probs = reweight_top_p(probs+eps, weight_top_p)

        probs = probs.clamp(eps, 1)
        c = D.Categorical(probs).sample((shape,))
        # move sample dimension first
        loc = loc.movedim(-1, 0).gather(0, c)
        scale = scale.movedim(-1, 0).gather(0, c)
        upper = cdfs[...,1,:].movedim(-1, 0).gather(0, c)
        lower = cdfs[...,0,:].movedim(-1, 0).gather(0, c)

        u = torch.rand(shape, *h.shape[:-1])
        # truncate
        u = u * (upper-lower) + lower

        u = u.clamp(eps, 1-eps)

        x = loc + bias - scale * (1/u - 1).log()

        if truncate_quantile is not None:
            x = x.sort(dim=0).values

            # print(truncate_quantile, list(x_.item() for x_ in x))

            idx = categorical_sample(
                x.new_zeros(x.shape[0]), 
                truncate_quantile=truncate_quantile)

            x = x[idx]
            
        x = x.clamp(self.lo, self.hi)
        x = x.clamp(*truncate)
        return x[0] if unwrap else x


class CensoredMixturePointyBoi(nn.Module):
    def __init__(self, n, res=1e-2, lo='-inf', hi='inf', sharp_bounds=(1e-5,2e3)):
        super().__init__()
        self.n = n
        self.res = res
        self.sharp_bounds = sharp_bounds
        # self.register_buffer('max_sharp', torch.tensor(float(max_sharp)))
        self.register_buffer('lo', torch.tensor(float(lo)))
        self.register_buffer('hi', torch.tensor(float(hi)))
        # TODO: init is not general-purpose
        self.bias = nn.Parameter(torch.cat((
            torch.zeros(n), torch.logspace(-3,1,n), torch.zeros(n)
            )))

    @property
    def n_params(self):
        return self.n*3

    def get_params(self, h):
        assert h.shape[-1] == self.n_params
        h = h+self.bias
        # get parameters fron unconstrained hidden state:
        logit_pi, loc, log_s = torch.chunk(h, 3, -1)
        # mixture coefficients
        log_pi = logit_pi - logit_pi.logsumexp(-1,keepdim=True)
        # location
        loc = loc.clamp(self.lo-10*self.res, self.hi+10*self.res)
        # sharpness
        # s = log_s.exp()
        # s = torch.min(F.softplus(log_s), self.max_sharp)
        s = F.softplus(log_s).clamp(*self.sharp_bounds)
        # s = log_s.exp().clamp(*self.sharp_bounds)
        return log_pi, loc, s

    def forward(self, h, x):
        """log prob of x under distribution parameterized by h"""
        log_pi, loc, s = self.get_params(h)    

        x = x.clamp(self.lo, self.hi)[...,None]
        xp, xm = x+self.res/2, x-self.res/2

        # numerical crimes follow

        # censoring
        lo_cens = x <= self.lo
        xm_ = torch.where(lo_cens, -h.new_ones([]), (xm-loc)*s)
        axm_ = torch.where(lo_cens, h.new_zeros([]), xm_.abs())
        hi_cens = x >= self.hi
        xp_ = torch.where(hi_cens, h.new_ones([]), (xp-loc)*s)
        axp_ = torch.where(hi_cens, h.new_zeros([]), xp_.abs())

        log_delta_cdf = (
            (xp_ - xm_ + xp_*axm_ - axp_*xm_).log() 
            - (axp_ + axm_ + axp_*axm_).log1p() 
            - math.log(2))
        
        # log prob
        r = {
            'log_prob': (log_pi + log_delta_cdf).logsumexp(-1)
        }
        # diagnostics
        with torch.no_grad():
            ent = D.Categorical(logits=log_pi).entropy()
            r |= {
                'min_sharpness': s.min(),
                'max_sharpness': s.max(),
                'min_entropy': ent.min(),
                'max_entropy': ent.max(),
                'marginal_entropy': D.Categorical(
                    log_pi.exp().mean(list(range(log_pi.ndim-1)))).entropy(),
                'min_loc': loc.min(),
                'max_loc': loc.max()
            }
        return r

    def cdf(self, h, x):
        log_pi, loc, s = self.get_params(h)  
        x_ = (x[...,None] - loc) * s 
        cdfs = x_ / (1+x_.abs()) * 0.5 + 0.5
        cdf = (cdfs * log_pi.softmax(-1)).sum(-1)
        return cdf


    def sample(self, h, shape=1):
        """
        Args:
            shape: additional sample shape to be prepended to dims
        """
        # if shape is None: shape = []

        log_pi, loc, s = self.get_params(h)
        c = D.Categorical(logits=log_pi).sample((shape,))
        # move sample dimension first
        loc = loc.movedim(-1, 0).gather(0, c)
        s = s.movedim(-1, 0).gather(0, c)

        u = torch.rand(shape, *h.shape[:-1])*2-1
        x_ = u / (1 - u.abs())
        x = x_ / s + loc

        return x.clamp(self.lo, self.hi)



"""
Notochord MIDI co-improviser server.
Notochord plays different instruments along with the player.
"""

# Authors:
#   Victor Shepardson
#   Intelligent Instruments Lab 2025

tui_doc = """Welcome to notochord homunculus.
You have 16 channels of MIDI, laid out in two rows of eight.
Each can be in one of three modes:
    input (-->02)
    follow (01->02)
    auto (02)
Input channels take MIDI input from the corresponding channel,
and send it to notochord. When using --thru, input is also copied
to the output.
    Follow channels act like a harmonizer; they query notochord for
a NoteOn whenever the followed channel has a NoteOn, and have the
corresponding NoteOff when the followed channel does.
    Auto channels are played autonomously by notochord.
Each channel can be muted independently of the others.
    Below the channel strips, there is a row of presets.
These can be edited in homunculus.toml; run `notochord files` to find 
it in your file explorer."""

# TODO: move soundfont / general MIDI stuff out of script

# TODO: make key bindings visibly click corresponding buttons
# TODO: make Mute a toggle but Reset a momentary

# TODO: color note by pitch class + register
# TODO: color instrument 1-128, MEL, DRUM, ANON, ANONDRUM
# TODO: color time
# TODO: id prediction as player / noto
# TODO: unify note log / prediction format
# TODO: grey out predictions when player or muted notochord

# TODO: controls display panel
# TODO: held notes display panel
# TODO: counts / weights display panel
# TODO: MIDI learn

import time
import random
import traceback
import shutil
from typing import Optional, Dict, Any
from numbers import Number
import math
import functools as ft
from pathlib import Path
from collections import defaultdict
from datetime import datetime

import toml_file
import mido
from tqdm import tqdm

import torch
torch.set_num_threads(1)

import iipyper, notochord
from notochord import Notochord, NotoPerformance, NoPossibleEvents
from notochord.util import deep_update
from iipyper import OSC, MIDI, run, Stopwatch, repeat, cleanup, TUI, profile, lock

from rich.panel import Panel
from rich.pretty import Pretty
from textual.reactive import reactive
from textual.widgets import Header, Footer, Static, Button, Log, RichLog, Label
from textual.screen import Screen, ModalScreen
from textual.containers import Grid

def now():
    return time.perf_counter()

def main(
    checkpoint="notochord-latest.ckpt", # Notochord checkpoint
    config:Dict[int,Dict[str,Any]]=None, # map MIDI channel : GM instrument
    preset:str=None,
    preset_file:Path=None,

    midi_prompt:Path=None,
    prompt_config:bool=True, # set channel config based on prompt
    prompt_channel_order:str=None,
    prompt_merge:bool=False,

    seed:int=0,

    initial_mute=False, # start with Notochord muted
    initial_query=None, # DEPRECATED, now inverse of `initial_stop`
    initial_stop=False, # if False, auto voices play immediately

    midi_in:Optional[str]=None, # MIDI port for player input
    midi_out:Optional[str]=None, # MIDI port for Notochord output
    midi_control:Optional[str]=None, # MIDI port for note messages as controls
    thru=False, # copy player input to output
    send_pc=False, # send program change messages
    dump_midi=False, # print all incoming MIDI
    suppress_midi_feedback=True,
    input_channel_map=None,

    stop_on_reset=False, # 
    stop_on_unmute=False, # 
    stop_on_end=False, # auto channels stop when end is sampled
    reset_on_end=False, # reset notochord when end is sampled
    end_exponent=1, # < 1 makes end more likely, >1 less likely
    end_thresh=None, # cumulative probability for deterministic end
    min_end_time=8, # prevent sampling end before this many seconds since reset

    balance_sample=False, # choose instruments which have played less recently
    n_recent=32, # number of recent note-on events to consider for above
    n_margin=8, # amount of 'slack' in the balance_sample calculation
    
    max_note_len=11, # in seconds, to auto-release stuck Notochord notes
    max_time=None, # max time between events
    nominal_time=True, #DEPRECATED

    min_vel=None, #mininum velocity (besides noteOffs)
    max_vel=None, # maximum velocity

    osc_port=None, # if supplied, listen for OSC to set controls on this port
    osc_host='', # leave this as empty string to get all traffic on the port

    punch_in=None, # DEPRECATED. now configured per-channel
    punch_out_after=None, # DEPRECATED. now configured per-channel

    use_tui=True, # run textual UI
    predict_input=True, # forecasted next events can be for input (preserves model distribution, but can lead to Notochord deciding not to play)
    predict_follow=False,
    debug_query=False, # don't query notochord when there is no pending event.
    testing=False,
    estimated_query_latency=10e-3,
    estimated_feed_latency=10e-3,
    lateness_margin=100e-3, # when events are playing later than this, slow down
    soft_lateness_margin=50e-3, # when events are playing later than this, slow down
    soundfont=None,
    limit_input=None,
    # thru_vel_offset=None,
    profiler=0,
    wipe_presets=False,
    ):
    """
    This a terminal app for using Notochord interactively with MIDI controllers and synthesizers. It allows combining both 'harmonizer' and 'improviser' features.
     
    Arguments to main can be given on the command line as flags, for example:

    `notochord homunculus --config "{1:{mode:auto, inst:1, mute:False}}" --initial-query --max-time 1`

    This says: play the grand piano autonomously on channel 1, start automatically, allow no more than 1 second between MIDI events. A different example:

    `notochord homunculus --config "{1:{mode:input, inst:1, mute:False}, 2:{mode:follow, inst:12, mute:False}}" --thru --send-pc`

    This says, take grand piano input on channel 1, and harmonize it with vibraphone on channel 2. Pass input through to the output, and also send program change messages to set the instruments on the synthesizer.

    You may also need to use the --midi-in or --midi-out flags to get MIDI to the right place.
    
    # MIDI Channels

    In homunculus, you have 16 MIDI channels. Each channel can be in one of three modes:

        * input (appearing like "-->01"), channel 1 comes from MIDI input channel 1.
        * follow (appearing like "01->02"), channel 2 plays whenever channel 1 plays.
        * auto (appearing like just "03"), channel 3 plays autonomously.

    Click the top section of each channel strip to cycle the mode.

    Each channel is also assigned a [General MIDI instrument](https://en.wikipedia.org/wiki/General_MIDI#Program_change_events). Each 'input' and 'auto' channel should have a unique General MIDI instrument, but 'follow' channels can be duplicates of others. If you try to assign duplicate instruments, they will automatically change to "anonymous" melodic or drum instruments, but still send the right program change messages when using --send-pc.

    Click the middle section of each channel strip to choose a new instrument.

    The bottom section of each channel strip allows muting individual voices.

    # Global controls

    Along the bottom, there are global query, sustain, mute and reset buttons. Query manually replaces the next pending note. Sustain stops all auto voices from playing without ending any open notes. Mute ends all open notes and stops auto voices. Reset ends all open notes, forgets all context and sets the Notochord model to its initial state.

    Args:
        checkpoint: path to notochord model checkpoint.

        config: 
            mapping from MIDI channels to voice specs.
            MIDI channels and General MIDI instruments are indexed from 1.
            see wikipedia.org/wiki/General_MIDI for values.
            There are 3 modes of voice, 'auto', 'follow' and 'input'. For example,
            ```
            {
                1:{
                    'mode':'input', 'inst':1
                }, # input grand piano on MIDI channel 1
                2:{
                    'mode':'follow', 'source':1, 'inst':1, 
                    'transpose':(-12,12)
                }, # harmonize the channel within 1 octave 1 with more piano
                3:{
                    'mode':'auto', 'inst':12, 'range':(36,72)
                }, # autonomous vibraphone voice in the MIDI pitch 36-72 range
                10:{
                    'mode':'auto', 'inst':129,
                }, # autonomous drums voice
                4:{
                    'mode':'follow', 'source':3, 'inst':10, 'range':(72,96)
                }, # harmonize channel 3 within upper registers of the glockenspiel
            }
            ```
            Notes:
            when two channels use the same instrument, one will be converted
            to an 'anonymous' instrument number (but it will still respect 
            the chosen instrument when using --send-pc)

            additional per-voice options:
            poly: pair of (min, max) simultaneous notes 
                (min should be 0 in most cases)
            duration: pair of (min, max) note duration
            punch_in: this causes the voice to switch between
                auto and input mode when input is received or not
            punch_out_after: time in seconds from last input NoteOn to revert   
                to auto mode

        preset: 
            preset name (in preset file) to load config from
        preset_file: 
            path to a TOML file containing channel presets
            the default config file is `homunculus.toml`; 
            running `notochord files` will show its location.

        midi_prompt:
            path to a MIDI file to read in as a prompt.
            note that prompts can alternatively be associated with presets 
            in `homunculus.toml`
        prompt_config:
            if True, set unmuted instruments to those in the prompt file.
            prompt config overrides presets,
            but config from the `--config flag` will override the prompt
        prompt_channel_order:
            method to re-order the channels in a MIDI prompt
            'channel' (default, leave as they are in file)
            'instrument': sort by instrument ID 
                (keeping associated anonymous IDs together)
            'instrument_drums': sort by instrument ID,
                but with drums starting on channel 10
            'notes': sort by most notes
        prompt_merge:
            if True, merge multiple channels with the same GM program into one
            Notochord voice (with overlapping notes dropped or shortened)
            if False, use anonymous instruments.

        seed:
            global random seed (added to preset random seed)

        initial_mute: 
            start 'auto' voices muted so it won't play with input.
        initial_stop: 
            if True, begin in a stopped state, 
            so 'auto' voices don't begin playing without input.

        midi_in: 
            MIDI ports for input. 
            default is to use all input ports.
            can be comma-separated list of ports.
        midi_out: 
            MIDI ports for output. 
            default is to use only virtual 'From iipyper' port.
            can be comma-separated list of ports.
        thru: 
            if True, copy input MIDI to output ports.
            only makes sense if input and output ports are different.
        send_pc: 
            if True, send MIDI program change messages to set the General MIDI
            instrument on each channel according to player_config and noto_config.
            useful when using a General MIDI synthesizer like fluidsynth.
        dump_midi: 
            if True, print all incoming MIDI for debugging purposes
        suppress_midi_feedback:
            attempt to allow use of the one loopback port for both input and
            output, by ignoring any MIDI input which is identical to an output
            within a few milliseconds.
            
        balance_sample:
            choose 'auto' voices which have played less recently,
            ensures that all configured instruments will play.
        n_recent: 
            number of recent note-on events to consider for above
        n_margin: 
            controls the amount of slack in the balance_sample calculation

        max_note_len: 
            time in seconds after which to force-release sustained 'auto' notes.
        max_time: 
            maximum seconds between predicted events for 'auto' voices.
            default is the Notochord model's maximum (usually 10 seconds).
        lateness_margin:
            when events are playing later than this (in seconds), slow down

        min_vel: mininum velocity (except for noteOffs where vel=0)
        max_vel: maximum velocity

        osc_port: 
            optional. if supplied, listen for OSC to set controls
        osc_host: 
            hostname or IP of OSC sender.
            leave this as empty string to get all traffic on the port

        use_tui: 
            run textual UI.
        predict_input: 
            forecasted next events can be for 'input' voices.
            generally should be True for manual input;
            use balance_sample to force 'auto' voices to play. 
            you might want it False if you have a very busy input.
        predict_follow:
            ditto for 'follow' voices. less obvious what the correct setting
            is, but False will be more efficient

        wipe_presets:
            if True, replaces your homunulus.toml with the defaults
            (may be useful after updating notochord)

        soundfont: path to a soundfont file from which default instrument 
            ranges will be loaded

        wipe_presets: CAUTION. replaces your homunculus.toml file with the 
            default. may fix errors after upgrading. backup your config file
            if you've been changing it! (`notochord files` to find it)
    """
    if osc_port is not None:
        osc = OSC(osc_host, osc_port)
    if midi_in is None and midi_control is not None:
        midi_in = set(mido.get_input_names()) - set(midi_control.split(','))
    midi = MIDI(midi_in, midi_out, suppress_feedback=suppress_midi_feedback)
    if midi_control is not None:
        midi_control = MIDI(
            midi_control, midi_control, 
            virtual_in_ports=0, virtual_out_ports=0)

    # backwards compat
    if initial_query is not None:
        initial_stop = not initial_query
    if punch_in is not None or punch_out_after is not None:
        raise ValueError("""
            punch_in and punch_out_after are now per-voice; 
            use --config flag or edit homunculus.toml
        """)

    estimated_latency = estimated_feed_latency + estimated_query_latency

    if input_channel_map is None: input_channel_map = {}

    ### Textual UI
    tui = NotoTUI()
    print = notochord.print = iipyper.print = tui.print
    ###

    if soundfont is None:
        sf_inst_ranges = {}
    else:
        # attempt to get instrument ranges from the soundfont
        # assumes first bank is used
        # not sure if entirely correct
        from sf2utils.sf2parse import Sf2File
        from sf2utils.generator import Sf2Gen

        with open(soundfont, 'rb') as file:
            soundfont = Sf2File(file)
        sf_presets = {
            p.preset:p
            for p in soundfont.presets 
            if hasattr(p,'bank') and p.bank==0}
        sf_drum_presets = {
            p.preset:p
            for p in soundfont.presets 
            if hasattr(p,'bank') and p.bank==128}

        def _get_range(i):
            if i>128:
                if i-128 in sf_drum_presets:
                    p = sf_drum_presets[i-128]
                else:
                    p = sf_drum_presets[0]
            else:
                if i-1 in sf_presets:
                    p = sf_presets[i-1]
                else:
                    p = sf_presets[0]
            # union of bags in preset
            # bag: intersection of range with union of bags in instrument
            preset_range = set()
            for b in p.bags:
                bag_range = set()
                if Sf2Gen.OPER_INSTRUMENT in b.gens: 
                    inst = soundfont.instruments[b[Sf2Gen.OPER_INSTRUMENT].amount]
                    for bi in inst.bags:
                        if Sf2Gen.OPER_SAMPLE_ID in bi.gens:
                            if bi.key_range is not None:
                                l,h = bi.key_range
                                bag_range |= set(range(l, h+1))
                if b.key_range is not None: 
                    # print(b.key_range)
                    l,h = b.key_range
                    bag_range &= set(range(l, h+1))
                preset_range |= bag_range
            # print(f'{preset_range=}')
            if len(preset_range):
                return min(preset_range), max(preset_range)
            return 0, 127
        sf_inst_ranges = {i:_get_range(i) for i in range(1,257)}
    def get_range(i):
        return sf_inst_ranges.get(i, (0,127))

    # print(f'{sf_inst_ranges=}')

    if not nominal_time:
        print('nominal_time is deprecated; nominal_time=False now sets lateness_margin to 0')
        lateness_margin = 0
    
    # make preset file if it doesn't exist
    cfg_dir = Notochord.user_data_dir()
    default_preset_file = cfg_dir / 'homunculus.toml'
    src_preset_file = Path(__file__).parent / 'homunculus.toml'
    if wipe_presets or not default_preset_file.exists():
        shutil.copy(src_preset_file, default_preset_file)

    ### presets and config
    try:    
        if preset_file is None:
            global_config = toml_file.Config(str(default_preset_file))
        else:
            global_config = toml_file.Config(str(preset_file))
        presets = global_config.get('preset', {})
        control_meta = global_config.get('control', {})
        action_meta = global_config.get('action', {})

    except Exception:
        print('WARNING: failed to load presets from file')
        print(traceback.print_exc(file=tui))
        global_config = {}
        presets = {}
        control_meta = []
        action_meta = []

    # print(f'{global_config=}')
    # print(f'{presets=}')

    # store control values
    controls = {ctrl['name']:ctrl.get('value', None) for ctrl in control_meta}

    # load notochord model
    try:
        noto = Notochord.from_checkpoint(checkpoint)
        noto.eval()
        noto.feed(0,0,0,0) # smoke test
        noto.query()
        noto.reset()
    except Exception:
        print("""error loading notochord model""")
        raise

    # defaults
    def default_config_channel(i):
        """default values for presets
        all fields should appear here, even if default is None
        """
        d = global_config.get('default', {})
        # toml_file's get method also sets the value for some reason??
        # and it refuses to set a None value...
        def get(k, default):
            return d[k] if k in d else default
        return {
            'mode': get('mode', 'auto'), 
            'inst': noto.first_anon_like(1),
            'mute': get('mute', True), 
            'source': max(1,i-1), 
            'note_shift': get('note_shift', 0),
            'duration': get('duration', None),
            'poly': get('poly', None),
            'range': get('range', None),
            'transpose': get('transpose', None),
            'cc': get('cc', toml_file.Config()),
            'punch_in': get('punch_in', False),
            'punch_out_after': get('punch_out_after', 0.25),
            }
    
    # convert MIDI channels to int
    # print(presets)
    for p in presets:
        if 'channel' not in p:
            p['channel'] = {}
        else:
            p['channel'] = {int(k):v for k,v in p['channel'].items()}   

    ### this feeds all events from the prompt file to notochord
    def do_prompt(prompt_file, channel_order=None, merge_channels=False, seed=None):
        prompt_random = random.Random(seed)
        prompt_file = Path(prompt_file).expanduser()
        if prompt_file.is_dir():
            prompt_file = prompt_random.choice([
                p for p in 
                prompt_file.glob('**/*')
                if p.is_file()
                ])
        print(f'{prompt_file=}')
        noto.reset()
        initial_state, inst_data = noto.prompt(
            prompt_file, merge=merge_channels)
        # first get the 16 most active parts
        insts = sorted(inst_data, key=lambda x: -inst_data[x].notes)[:16]
        print(f'{inst_data=}')
        # then order them onto channels
        def make_cfg(i):
            return {
                'inst':i, 'mute':False, 
                'program_inst':inst_data[i].orig_inst}
        if channel_order=='instrument':
            # sort by (original_inst, noto_inst)
            config = {
                c+1:make_cfg(i)
                for c,i in enumerate(sorted(
                    insts, key=lambda k: (inst_data[k].orig_inst, k)))}
        elif channel_order=='instrument_drums':
            insts = sorted(insts, key=lambda k: (inst_data[k].orig_inst, k))
            drum_insts = [i for i in insts if noto.is_drum(i)]
            mel_insts = [i for i in insts if i not in drum_insts]
            config = {c+10:make_cfg(i) for c,i in enumerate(drum_insts)}
            cs = sorted(set(range(1,17)) - set(config))
            config.update({c:make_cfg(i) for c,i in zip(cs, mel_insts)})
        elif channel_order=='main_keyboard_first':
            drum_insts = [i for i in insts if noto.is_drum(i)]
            kb_insts = []
            nonkb_insts = []
            for i in insts:
                if i in drum_insts: continue
                l,h = get_range(inst_data[i].orig_inst)
                if l<=21 and h>=108:
                    kb_insts.append(i)
                else:
                    nonkb_insts.append(i)
            mel_insts = kb_insts + nonkb_insts
            config = {c+10:make_cfg(i) for c,i in enumerate(drum_insts)}
            cs = sorted(set(range(1,17)) - set(config))
            config.update({c:make_cfg(i) for c,i in zip(cs, mel_insts)})
        elif channel_order=='notes':
            config = {c+1:make_cfg(i) for c,i in enumerate(insts)}
        elif isinstance(channel_order, list):
            config = {
                channel_order[min(inst_data[i].channels)]:make_cfg(i) 
                for i in reversed(insts)}
        elif channel_order=='channel' or channel_order is None:
            # NOTE this takes the minimum channel if an instrument
            # appears on multiple channels
            # the instrument with most notes will take precedence
            # if multiple instruments appear
            # but could be better to avoid collisions in some cases...
            config = {
                min(inst_data[i].channels)+1:make_cfg(i) 
                for i in reversed(insts)}
        else:
            raise ValueError(f"""
                unknown {channel_order=} 
                (options are 'instrument', 'notes', or 'channel')""")
        return initial_state, config
            
    global_initial_state = None
    config_ingest = None
    if midi_prompt is not None:
        global_initial_state, config_ingest = do_prompt(
            midi_prompt, 
            merge_channels=prompt_merge, channel_order=prompt_channel_order)
        if not prompt_config:
            config_ingest = None
    initial_state = global_initial_state

    # process prompts in each preset
    for p in presets:
        if 'prompt' in p:
            print(f'prompting "{p["prompt"]}" for preset "{p["name"]}"')
            prompt_seed = p.get('seed')
            if seed is None and p.get('seed') is None:
                prompt_seed = None
            else:
                prompt_seed = (seed or 0) + p.get('seed', 0)
            p['initial_state'], prompt_cfg = do_prompt(
                p['prompt'], 
                merge_channels=p.get('prompt_merge_channels'),
                channel_order=p.get('prompt_channel_order'),
                seed=prompt_seed
                )
            # print(f'{prompt_cfg=}')
            preset_cfg = p['channel']
            for k,chan_cfg in prompt_cfg.items():
                chan_cfg.update(preset_cfg.get(k, {}))
                preset_cfg[k] = chan_cfg
            # print(f"preset channel config: {preset_cfg}")

    # def validate_config():
    #     assert all(
    #         v['source'] in config for v in config.values() if v['mode']=='follow'
    #         ), 'ERROR: no source given for follow voice'
    #     # TODO: check for follow cycles
    # validate_config()

    def mode_insts(t, allow_muted=True):
        if isinstance(t, str):
            t = t,
        # set of instruments with given mode(s)
        return {
            v['inst'] for v in config.values() 
            if v['mode'] in t and (allow_muted or not v['mute'])
            }
    def mode_chans(t):
        if isinstance(t, str):
            t = t,
        # list of channels with given mode
        return [k for k,v in config.items() if v['mode'] in t]
    def channel_inst(c):
        return config[c]['inst']
    # def channel_insts():
        # list of channel,instrument pairs
        # return [(c,channel_inst(c)) for c in config]
    def inst_ranges(insts):
        # instruments to sets of allowed MIDI numbers
        r = {}
        for v in config.values():
            i = v['inst']
            if i in insts:
                lo, hi = v.get('range') or get_range(i) # inclusive range
                s = set(range(lo, hi+1))
                if i in r:
                    r[i] |= s
                else:
                    r[i] = s
        return r
    def auto_inst_channel(i):
        for k,v in config.items():
            if v['mode']=='auto' and v['inst']==i:
                return k
        return None
    def channel_followers(chan):
        # print(f'channel_followers {chan=} {config=}')
        # return channel of all 'follow' voices with given source
        return [
            k for k,v in config.items() 
            if v['mode']=='follow' 
            and v.get('source', None)==chan]

    def get_free_anon_like(i):
        s = set(noto.anon_like(i)) - {d['inst'] for d in config.values()}
        return next(iter(s))

    def dedup_inst(c, i):
        if any(
            c_other!=c
            and d['inst']==i 
            and d['mode']!='follow'
            for c_other, d in config.items()
            ):
            i = get_free_anon_like(i)
        return i

    def do_send_pc(c, i):
        # warn_inst(i)
        # assuming fluidsynth -o synth.midi-bank-select=mma
        if noto.is_drum(i):
            midi.control_change(channel=c-1, control=0, value=1)
        else:
            midi.control_change(channel=c-1, control=32, value=0)
            midi.control_change(channel=c-1, control=0, value=0)
        if noto.is_anon(i):
            program = 0
        else:
            # convert to 0-index
            program = (i-1) % 128
        midi.program_change(channel=c-1, program=program)

    # simple class to hold pending event prediction
    class Prediction:
        def __init__(self):
            self.gate = not initial_mute
            self.reset()
        def reset(self):
            self.stopped = False
            self.last_event_time = None
            self.lateness = 0
            # self.on_cum_end_prob = 0
            # self.off_cum_end_prob = 0
            self.cum_end_prob = 0
            self.clear()
        def clear(self):
            self.event = None
            self.next_event_time = None
            # tui.defer(prediction=None)
        def occurred(self):
            self.last_event_time = self.next_event_time
            self.clear()
        def set(self, event):
            if event.get('vel',0) > 0:
                self.stopped = False
            if event.get('time') is None:
                event['time'] = self.time_since()
            self.event = event
            self.next_event_time = event['time'] + (
                self.last_event_time or now())
        def time_since(self):
            """current actual time since nominal time of last played event"""
            if self.last_event_time is None:
                return 0
            else:
                return now() - self.last_event_time
        def time_until(self):
            if self.next_event_time is None: 
                return float('inf')
            else:
                return self.next_event_time - now()
        def is_auto(self):
            if self.event is None:
                return False
            return self.event['inst'] in mode_insts('auto')
        def sample_end(self):
            if self.event is None:
                return False
            
            end_prob = self.event.get('end', 0) ** end_exponent
            self.cum_end_prob = 1 - (1-self.cum_end_prob)*(1-end_prob)
            # print(f'{self.cum_end_prob=}')

            if self.event['vel']>0:
                # self.on_cum_end_prob = 1 - (1-self.on_cum_end_prob)*(1-end_prob)
                # print(f'{self.on_cum_end_prob=}')
                return False
            # else:
            #     self.off_cum_end_prob = 1 - (1-self.off_cum_end_prob)*(1-end_prob)
            #     print(f'{self.off_cum_end_prob=}')
            # # cumulative end probability
            if end_thresh is None:
                return random.random() < end_prob
            else:
                return self.cum_end_prob > end_thresh
    pending = Prediction()    

    # tracks held notes, recently played instruments, etc
    # NOTE: Notochord now tracks held notes,
    # but NotoPerformance also tracks channels,
    # lets you attach user data to notes,
    # remembers past events, and contains other utilities
    history = NotoPerformance()

    status = {'reset_time':now()}

    action_queue = []

    def display_event(tag, memo, inst, pitch, vel, channel, **kw):
        """print an event to the terminal"""
        if tag is None:
            return
        now = str(datetime.now())[:-2]
        # tstr = "O" if kw["time"]<=1e-3 else ""
        s = f'{now}   inst {inst:3d}   pitch {pitch:3d}   vel {vel:3d}   ch {channel:2d} {tag}'
        # s = f'{tag}:\t{inst=:4d}   {pitch=:4d}   {vel=:4d}   {channel=:3d}'
        if memo is not None:
            s += f' ({memo})'
        tui.defer(note=s)

    # @profile(print=print, enable=profiler)
    def send_midi(note, velocity, channel):
        kind = 'note_on' if velocity > 0 else 'note_off'
        cfg = config[channel]
        # get channel note map
        note_map = cfg.get('note_map', {})
        if note in note_map:
            note = note_map[note]
        elif 'note_shift' in cfg:
            note = note - cfg['note_shift']

        if note < 0:
            print(f'WARNING: dropped note < 0 ({note}, {channel=})')
            return
        if note >= 128:
            print(f'WARNING: dropped note >= 128 ({note}, {channel=})')
            return

        port = cfg.get('port', None)

        midi.send(kind, note=note, velocity=velocity, channel=channel-1, port=port)

    @profile(print=print, enable=profiler)
    def play_event(
            channel, 
            parent=None, # parent note as (channel, inst, pitch)
            feed=True, 
            send=True, 
            tag=None, memo=None):
        """realize an event as MIDI, terminal display, and Notochord update"""
        with profile('rest', print=print, enable=profiler):
            event = pending.event
            if event is None:
                print("WARNING: play_event on null event")
                return
            time_until = pending.time_until()
            pending.lateness = -time_until # set at the time of playing
            # if time_until < 10e-3:
                # print(f'late {-time_until} at {now()=}')
            pending.occurred()
            # normalize values
            vel = event['vel'] = math.ceil(event['vel'])

            # send out as MIDI
            if send:
                send_midi(event['pitch'], vel, channel)

            # print
            display_event(tag, memo=memo, channel=channel, **event)

        if feed:
            # feed to NotoPerformance
            # put a stopwatch in the held_note_data field for tracking note length
            with profile('history.feed', print=print, enable=profiler>1):
                history.feed(held_note_data={
                    'duration':Stopwatch(),
                    'parent':parent
                    }, channel=channel, **event)
            # feed to model
            with profile('noto.feed', print=print, enable=profiler):
                noto.feed(**event)

        follow_event(event, channel, feed=feed)

    @profile(print=print, enable=profiler)
    def follow_event(source_event, source_channel, feed=True):
        source_vel = source_event['vel']
        source_pitch = source_event['pitch']
        source_inst = source_event['inst']
        source_k = (source_channel, source_inst, source_pitch)

        dt = 0

        if source_vel > 0:

            # NoteOn
            for noto_channel in channel_followers(source_channel):
                cfg = config[noto_channel]
                
                if cfg.get('mute', False): continue

                noto_inst = cfg['inst']
                min_x, max_x = cfg.get('transpose') or (-128,128)
                lo, hi = cfg.get('range') or (0,127)

                already_playing = {
                    p for i,p in noto.held_notes if noto_inst==i}
                # already_playing = {
                #     note.pitch for note in history.notes if noto_inst==note.inst}
                # print(f'{already_playing=}')

                pitch_range = range(
                    max(lo,source_pitch+min_x), min(hi, source_pitch+max_x+1))
                pitches = (
                    set(pitch_range) - {source_pitch} - already_playing
                )

                if len(pitches)==0:
                    # edge case: no possible pitch
                    print(f'skipping follow {noto_channel=}, no pitches available')
                    print(f'{pitch_range} minus {source_pitch} minus {already_playing}')
                    print(f'{lo=} {min_x=} {hi=} {max_x=} {source_pitch=}')
                    continue
                elif len(pitches)==1:
                    # edge case: there is exactly one possible pitch
                    pending.set(dict(
                        inst=noto_inst, pitch=list(pitches)[0], 
                        time=dt, vel=source_vel))
                else:
                    # notochord chooses pitch
                    with profile('noto.query', print=print, enable=profiler):
                        pending.set(noto.query(
                            next_inst=noto_inst, next_time=dt, next_vel=source_vel,
                            include_pitch=pitches))
                        
                pending.sample_end() # track cumulative end prob
                play_event(
                    noto_channel, feed=feed,
                    parent=source_k, tag='NOTO', memo='follow')
        # NoteOff
        else:
            # print(f'{history.note_data=}')
            dependents = [
                noto_k # chan, inst, pitch
                for noto_k, note_data
                in history.note_data.items()
                if note_data['parent']==source_k
            ]

            for noto_channel, noto_inst, noto_pitch in dependents:
                pending.set(dict(
                    inst=noto_inst, pitch=noto_pitch, time=dt, vel=0))
                play_event(noto_channel, feed=feed, tag='NOTO', memo='follow')

    # @lock
    @profile(print=print, enable=profiler)
    def noto_reset():
        """reset Notochord and end all of its held notes"""
        print('RESET')

        status['reset_time'] = now()

        # cancel pending predictions
        pending.reset()
        
        # end Notochord held notes
        # skip feeding for speed, since notochord gets reset anyway
        end_held(feed=False, memo='reset')
                
        # reset notochord state
        noto.reset(state=initial_state)
        # reset history
        history.push()

        if stop_on_reset:
            noto_stop()

        # TODO: feed note-ons from any held input/follower notes?

    # @lock
    @profile(print=print, enable=profiler)
    def noto_mute(sustain=False):
        tui.query_one('#mute').label = 'UNMUTE' if pending.gate else 'MUTE'
        # if sustain:
        tui.query_one('#sustain').label = 'END SUSTAIN' if pending.gate else 'SUSTAIN'

        pending.gate = not pending.gate

        if sustain:
            print('END SUSTAIN' if pending.gate else 'SUSTAIN')
        else:
            print('UNMUTE' if pending.gate else 'MUTE')

        # if unmuting, we're done
        if not pending.gate:
            # cancel pending predictions
            if not sustain:
                end_held(memo='mute')
            pending.clear()
        else:
            if stop_on_unmute:
                noto_stop()

    def noto_stop():
        print('STOP')
        end_held(memo='stop on end')
        pending.stopped = True
        
    @profile(print=print, enable=profiler)    
    def end_held(feed=True, channel=None, memo=None):
        # end+feed all held notes
        channels = mode_chans('auto') if channel is None else [channel]
        ended_any = False
        for note in history.notes:
            if note.chan in channels:
                pending.set(dict(inst=note.inst, pitch=note.pitch, vel=0))
                play_event(
                    channel=note.chan, feed=feed, tag='NOTO', memo=memo)
                ended_any = True
        return ended_any

    # query Notochord for a new next event
    # @lock
    @profile(print=print, enable=profiler)
    def auto_query(
            predict_input=predict_input, 
            predict_follow=predict_follow,
            immediate=False):
        
        # NOTE: replaced this with duration constraints;
        # should test more before deleting
        # check for stuck notes
        # and prioritize ending those
        # for (_, inst, pitch), note_data in history.note_data.items():
        #     dur = note_data['duration'].read()
        #     if (
        #         inst in mode_insts('auto') 
        #         and dur > max_note_len*(.1+controls.get('steer_duration', 1))
        #         ):
        #         # query for the end of a note with flexible timing
        #         # with profile('query', print=print, enable=profiler):
        #         t = pending.time_since()
        #         mt = max(t, min(max_time or np.inf, t+0.2))
        #         pending.set(noto.query(
        #             next_inst=inst, next_pitch=pitch,
        #             next_vel=0, min_time=t, max_time=mt))
        #         print(f'END STUCK NOTE {inst=},{pitch=},{dur=}')
        #         return

        if immediate:
            # sampling immediately after realizing an event
            min_time = 0
        else:
            if pending.event is not None and not pending.is_auto():
                # re-sampling ahead after sampling an input or follow voice
                min_time = pending.event['time']
            else:
                # otherwise sampling after some delay
                min_time = pending.time_since()
            min_time = min_time-estimated_feed_latency
        # print(f'{immediate=} {min_time=}')

        # print(f'{pending.lateness=} {pending.time_since()=}')

        lateness = pending.lateness if immediate else -pending.time_until()
        if lateness > lateness_margin:
            backoff = lateness + estimated_latency
            if backoff > min_time:
                min_time = backoff
                print(f'set {min_time=} due to {lateness=}')

        # if max_time < min_time, exclude input instruments
        input_late = max_time is not None and max_time < min_time

        inst_modes = ['auto']
        if predict_follow:
            inst_modes.append('follow')
        if predict_input and not input_late:
            inst_modes.append('input')
        allowed_insts = mode_insts(inst_modes, allow_muted=False)

        # held_notes = history.held_inst_pitch_map()
        # print(f'{held_notes=}')
        steer_time = 1-controls.get('steer_rate', 0.5)
        steer_pitch = controls.get('steer_pitch', 0.5)
        steer_density = controls.get('steer_density', 0.5)
        steer_velocity = controls.get('steer_velocity', 0.5)

        if lateness > soft_lateness_margin and lateness <= lateness_margin:
            steer_time = 1
            print(f'set {steer_time=} due to {lateness=}')

        rhythm_temp = controls.get('rhythm_temp', 1)
        timing_temp = controls.get('timing_temp', 1)
        
        tqt = (max(0,steer_time-0.5), min(1, steer_time+0.5))
        tqp = (max(0,steer_pitch-0.5), min(1, steer_pitch+0.5))
        tqv = (max(0,steer_velocity-0.5), min(1, steer_velocity+0.5))

        # idea: maintain an 'instrument presence' quantity
        # incorporating time since / number of notes playing
        # ideally this would distinguish sustained from percussive instruments too
        
        # balance_sample: note-ons only from instruments which have played less
        inst_weights = None
        if balance_sample:
            counts = defaultdict(int)
            for i,c in history.inst_counts(n=n_recent).items():
                counts[i] = c
            # print(f'{counts=}')
            inst_weights = {}
            mc = max(counts.values()) if len(counts) else 0
            for i in allowed_insts:
                # don't upweight player controlled instruments though
                if i in mode_insts('auto'):
                    inst_weights[i] = math.exp(max(0, mc - counts[i] - n_margin))
                else:
                    inst_weights[i] = 1.
        # print(f'{inst_weights=}')

        # VTIP is better for time interventions,
        # VIPT is better for instrument interventions
        if min_time > estimated_latency or abs(steer_time-0.5) > abs(steer_pitch-0.5):
            query_method = noto.query_vtip
            # print('VTIP')
        else:
            query_method = noto.query_vipt
            # print('VIPT')
        # query_method = noto.query_vipt ### DEBUG
        # query_method = noto.query_vtip ### DEBUG
        query_method = profile(print=print, enable=profiler)(query_method)

        # print(f'considering {insts} for note_on')
        # use only currently selected instruments
        inst_pitch_map = inst_ranges(allowed_insts)
        note_on_map = {
            i: set(inst_pitch_map[i])#-set(held_notes[i]) # exclude held notes
            for i in allowed_insts#allowed_insts
            if i in inst_pitch_map
        }

        min_polyphony = {}
        max_polyphony = {}
        min_duration = {}
        max_duration = {}
        for i in note_on_map:
            c = auto_inst_channel(i)
            if c is None: continue
            cfg = config[c]
            if cfg is not None and cfg.get('poly') is not None:
                min_polyphony[i], max_polyphony[i] = cfg['poly']
            if cfg is not None and cfg.get('duration') is not None:
                min_duration[i], max_duration[i] = cfg['duration']
            elif max_note_len is not None:
                max_duration[i] = max_note_len

        # print(note_on_map, note_off_map)

        max_t = None if max_time is None else max(max_time, min_time+0.2)

        # print(f'{note_on_map=}')
        try:
            pending.set(query_method(
                note_on_map, #note_off_map,
                min_polyphony=min_polyphony, max_polyphony=max_polyphony,
                min_duration=min_duration, max_duration=max_duration,
                min_time=min_time, max_time=max_t,
                min_vel=min_vel, max_vel=max_vel,
                truncate_quantile_time=tqt,
                truncate_quantile_pitch=tqp,
                truncate_quantile_vel=tqv,
                rhythm_temp=rhythm_temp,
                timing_temp=timing_temp,
                steer_density=steer_density,
                inst_weights=inst_weights,
                no_steer=mode_insts(('input','follow'), allow_muted=False),
            ))
        except NoPossibleEvents:
            pass
            # print(f'stopping; no possible events')
            # pending.stopped = True
            # pending.clear()
        except Exception:
            print(f'WARNING: query failed. {allowed_insts=} {note_on_map=}')
            print(f'{noto.held_notes=}')
            print(f'{config=}')
            traceback.print_exc(file=tui)
            # pending.clear()

    #### MIDI handling

    # print all incoming MIDI for debugging
    if dump_midi:
        @midi.handle
        def _(msg):
            print(msg)

    @midi.handle(type='program_change')
    def _(msg):
        """
        TODO:Program change events set GM instruments on the corresponding channel
        """
        # c = msg.channel+1
        # c = input_channel_map.get(c, c)
        # i = msg.program+1
        # action_queue.append(ft.partial(set_inst(c,i)))
        # if thru:
        raise NotImplementedError

    @midi.handle(type='pitchwheel')
    def _(msg):
        """
        pitchwheel affects steer_pitch
        """
        if thru:
            midi.send(msg)
        # print(msg)
        controls['steer_pitch'] = (msg.pitch+8192)/16384
        # print(controls)
        # print(f'{controls["steer_pitch"]=}')

    # very basic CC handling for controls
    control_cc = {}
    control_osc = {}
    control_note = {} # note number : (control, value)
    for ctrl in control_meta:
        name = ctrl['name']
        ccs = ctrl.get('control_change', [])
        note_values = ctrl.get('note_value', {})
        if isinstance(ccs, Number) or isinstance(ccs, str):
            ccs = (ccs,)
        for cc in ccs:
            control_cc[cc] = ctrl
        for note, value in note_values.items():
            control_note[int(note)] = (ctrl, value)
        control_osc[f'/notochord/homunculus/{name}'] = ctrl
    # print(f'{control_note=}')
    action_cc = {}
    action_note = {}
    action_osc = {}
    for act in action_meta:
        name = act['name']
        ccs = act.get('control_change', [])
        notes = act.get('note', [])
        if isinstance(ccs, Number) or isinstance(ccs, str):
            ccs = (ccs,)
        if isinstance(notes, Number) or isinstance(notes, str):
            notes = (notes,)
        for cc in ccs:
            action_cc[cc] = act
        for note in notes:
            action_note[note] = act
        action_osc[f'/notochord/homunculus/{name}'] = act

    def dispatch_action(k, v=None):   
        if k=='reset':
            action_queue.append(noto_reset)
        elif k=='query':
            action_queue.append(auto_query)
        elif k=='mute':
            action_queue.append(noto_mute)
        elif k=='preset':
            action_queue.append(ft.partial(set_preset, v or 0))
        elif k=='preset_reset':
            action_queue.append(ft.partial(set_preset, v or 0))
            action_queue.append(noto_reset)
        elif k=='stop':
            action_queue.append(noto_stop)
        else:
            print(f'WARNING: action "{k}" not recognized')

    @midi.handle(type='control_change')
    def _(msg):
        """
        these are global controls listening on all channels: 
        CC 01: steer pitch (>64 higher pitches, <64 lower)
        CC 02: steer density (>64 more simultaneous notes, <64 fewer)
        CC 03: steer rate (>64 more events, <64 fewer)
        """
        if msg.control in control_cc:
            ctrl = control_cc[msg.control]
            name = ctrl['name']
            lo, hi = ctrl.get('range', (0,1))
            controls[name] = msg.value/127 * (hi-lo) + lo
            print(f"{name}={controls[name]}")

        if msg.control in action_cc:
            k = action_cc[msg.control]['name']
            v = action_cc[msg.control].get('value')
            dispatch_action(k, v)

    if midi_control is not None:
        preset_keys = range(112,120)
        momentary_keys = [120]
        # TODO: add toggle keys
        for note in preset_keys:
            midi_control.note_on(
                channel=0, note=note, velocity=0)
            
        @midi_control.handle(type=('note_on'))
        def _(msg, port):
            print('control note event', msg)

            if msg.velocity and msg.note in control_note:
                ctrl, value = control_note[msg.note]
                name = ctrl['name']
                # lo, hi = ctrl.get('range', (0,1)) # TODO
                controls[name] = value
                print(f"{name}={controls[name]}")

            if msg.velocity and msg.note in action_note:
                k = action_note[msg.note]['name']
                v = action_note[msg.note].get('value')
                dispatch_action(k, v)
            
            # MIDI feedback (lights)
            if msg.note in preset_keys:
                # if msg.velocity==0: return
                if msg.velocity>0:
                    for note in preset_keys:
                        midi_control.note_on(
                            channel=0, note=note, velocity=0, port=port)
                midi_control.note_on(
                    channel=0, note=msg.note, velocity=70 if msg.velocity else 127, port=port)
            if msg.note in momentary_keys:
                midi_control.note_on(
                    channel=0, note=msg.note, velocity=msg.velocity, port=port)

    if osc_port is not None:
        @osc.handle('/notochord/homunculus/*')
        def _(route, *a):
            # print('OSC:', route, *a)
            if route in control_osc:
                ctrl = control_osc[route]
                name = ctrl['name']
                assert len(a)==1
                arg = a[0]
                assert isinstance(arg, Number)
                lo, hi = ctrl.get('range', (0,1))   
                controls[name] = min(hi, max(lo, arg))
                print(f"{name}={controls[name]}")

            if route in action_osc:
                action = action_osc[route]
                k = action['name']
                dispatch_action(k, v=a[0] if len(a) else action.get('value'))

    input_sw = Stopwatch()
    dropped = set()# (channel, pitch)
    input_dts = []
    @midi.handle(type=('note_on', 'note_off'))
    def _(msg):
        """
        MIDI NoteOn and NoteOff events affect input channels
        e.g. a channel displaying -->01 will listen to note events on channel 1
        a channel displaying 02->03 will follow note events on channel 2
        """
        channel = msg.channel + 1
        # convert from 0-index
        channel = input_channel_map.get(channel, channel)
        cfg = config[channel]

        if cfg['punch_in']:
            # set_mode(channel, 'input')
            # print(f'punch {channel=} {cfg=}')
            action_queue.append(ft.partial(set_mode, channel, 'input'))

        if not (cfg['punch_in'] or channel in mode_chans('input')):
            # print(f"{channel=} {mode_chans('input')=}")
            print(f'WARNING: ignoring MIDI {msg} on non-input channel {channel}')
            return
        
        if cfg['mute']:
            print(f'WARNING: ignoring MIDI {msg} on muted channel {channel}')
            return

        inst = channel_inst(channel)
        pitch = msg.note + cfg.get('note_shift', 0)
        vel = msg.velocity if msg.type=='note_on' else 0
        
        dt = input_sw.punch()
        # print(f'EVENT {dt=} {msg}')
        if len(input_dts) >= 10:
            input_dts.pop(0)
        input_dts.append(dt)
        input_dens = len(input_dts) / sum(input_dts)
        # TODO: 
        # want to drop input when event density is high,
        # not just dt is short
        k = (channel,pitch)
        if vel==0 and k in dropped:
            dropped.remove(k)
            print(f'WARNING: ignoring rate-limited input')
            return
        if vel>0 and limit_input and input_dens>limit_input:
            print(f'WARNING: ignoring rate-limited input {input_dens=}')
            dropped.add(k)
            return 
        
        action_queue.append(ft.partial(
            do_note_input, channel, inst, pitch, vel))

    def do_note_input(channel, inst, pitch, vel):
        # feed event to Notochord
        pending.set({'inst':inst, 'pitch':pitch, 'vel':vel})
        pending.sample_end() # track cumulative end prob
        play_event(channel=channel, send=thru, tag='PLAYER')

    @profile(print=print, enable=profiler)
    def auto_event():
        # print('auto_event')
        # 'auto' event happens:
        event = pending.event
        inst, pitch, vel = event['inst'], event['pitch'], math.ceil(event['vel'])
        chan = auto_inst_channel(inst)
        if chan is None:
            raise ValueError(f"channel not found for instrument {inst}")

        # shouldn't happen, but prevent
        # note on which is already playing or note off which is not
        if (vel>0) == ((inst, pitch) in noto.held_notes): 
            print(f'WARNING: re-query for invalid {vel=}, {inst=}, {pitch=}')
            auto_query()
            return
                
        do_stop = False
        do_reset = False
        if (
            (stop_on_end or reset_on_end) 
            and (now()-status['reset_time'] > min_end_time) 
            and pending.sample_end()
            ):
            print('END')
            do_stop = stop_on_end
            do_reset = reset_on_end # needs to happen after final event plays

        play_event(channel=chan, tag='NOTO')

        if do_reset:
            noto_reset()
        elif do_stop:
            noto_stop()

    # @profile(print=print, enable=profiler)
    def maybe_punch_out():
        if not any(cfg['punch_in'] for cfg in config.values()):
            return

        none_held = set(mode_chans('input')) - {c for c,_,_ in history.notes}

        for c in none_held:
            if config[c]['punch_in']:
                after = config[c]['punch_out_after']
                t = history.last_event_time_ns(channel=c, on=True)
                if t is not None and t < time.time_ns() - after*1e9:
                    set_mode(c, 'auto')

    @cleanup
    def _():
        """end any remaining notes"""
        print(f'cleanup: {noto.held_notes=}')
        # TODO: this should run in repeat thread..?
        end_held(feed=False, memo='cleanup')

    ### update_* keeps the UI in sync with the state

    def update_config():
        # pass
        # print(config)
        for c,v in config.items():
            # tui.set_channel(c, v)
            tui.call_from_anywhere(tui.set_channel, c, v)

    def update_presets():
        for k,p in enumerate(presets):
            # tui.set_preset(k, p.get('name'))
            tui.call_from_anywhere(tui.set_preset, k, p.get('name'))

    ### set_* does whatever necessary to change channel properties
    ### calls update_config() to keep the UI in sync

    def set_mode(c, m, update=True):
        if c in config:
            prev_m = config[c]['mode']
        else:
            prev_m = None
        if m==prev_m:
            return
        
        if m=='follow':
            if 'source' not in config[c]:
                print('WARNING: follower without a source, setting to 1')
                config[c]['source'] = 1
        
        config[c]['mode'] = m
        print(f'set channel {c} from {prev_m} to {m} mode')

        ### NOTE changing a channel with followers to input causes stuck notes

        if prev_m=='follow':
            # emancipate held notes
            for (dep_c,_,_), note_data in history.note_data.items():
                if dep_c==c:
                    note_data['parent'] = None

        if prev_m=='auto':
            # release held notes
            end_held(channel=c, memo='mode change')

        if m=='auto':
            # immediately refresh prediction
            pending.clear()

        if update:
            update_config()

    def set_inst(c, i, program=None, update=True, allow_pc=True):
        # print(f'set channel {c} instrument {i}')
        req_i = i
        if c in config:
            prev_i = config[c]['inst']
        else:
            prev_i = None

        # don't steal from lower channels
        lower_insts = {d['inst'] for c2, d in config.items() if c2 < c}
        if i in lower_insts:
            # first anon not used by a lower channel
            i = min(set(noto.anon_like(i)) - lower_insts)

        # end held notes on old instrument
        if prev_i!=i and config[c]['mode']!='input':
            end_held(channel=c, memo='instrument change')
            pending.clear()

        # then set config
        config[c]['inst'] = i

        # steal from higher channels
        for c2 in range(c+1,17):
            i2 = config[c2]['inst']
            if i2==i or i2 in lower_insts:
                set_inst(
                    c2, get_free_anon_like(i), 
                    update=False, allow_pc=False)

        # send pc if appropriate
        if send_pc and allow_pc:
            do_send_pc(c, req_i if program is None else program)

        print(f'set channel {c} to instrument {i} (was {prev_i}, requested {req_i})')
        
        if update:
            update_config()

    # @lock
    def set_mute(c, b, update=True):
        config[c]['mute'] = b
        if b:
            print(f'mute channel {c}')
            # release held notes
            if config[c]['mode']!='input':
                end_held(channel=c, memo='mute channel')
                pending.clear()
        else:
            print(f'unmute channel {c}')

        if update:
            update_config()

    def set_cc(c, ccs, update=True):
        # print(f'{c=}, {ccs=}')
        for k,v in ccs.items():
            k = int(k)
            midi.send('control_change', channel=c-1, control=k, value=v)
            # print(f'{(k,v)=}')

    # @lock
    def set_preset(p, update=True):
        nonlocal initial_state
        if p >= len(presets): return
        print(f'load preset: {p+1}')
        preset = presets[p]

        overlay = preset.get('overlay', False)

        state = preset.get('initial_state')
        if state is None:
            if overlay:
                print(f'leaving current initial state')
            else:
                print(f'using global initial state')
                initial_state = global_initial_state
        else:
            print(f'using initial state from preset prompt')
            initial_state = state

        if update:
            for i in range(NotoPresets.n_presets):
                tui.query_one('#'+preset_id(i)).variant = (
                    'warning' if i==p else 'default')

        set_config(preset['channel'], overlay=overlay, update=update)

    def set_config(cfg, overlay=False, update=True):
        # NOTE: config should *not* be updated before calling set_* 
        # TODO: need to force notes off if note_shift changes
        if overlay:
            # just apply settings which are in the preset
            for c in sorted(cfg):
                v = {**cfg.get(c, {})}
                if 'mode' in v:
                    set_mode(c, v.pop('mode'), update=False)
                if 'inst' in v:
                    set_inst(
                        c, v.pop('inst'), 
                        program=v.pop('program_inst', None), 
                        update=False)
                if 'mute' in v:
                    set_mute(c, v.pop('mute'), update=False)
                if 'cc' in v:
                    set_cc(c, v.pop('cc'), update=False)
                config[c].update(v)
        else:
            # deterministically apply preset on top of defaults
            # for c in range(16,0,-1):    
            for c in range(1,17):    
                v = default_config_channel(c)
                # v.update(cfg.get(c, {}))
                deep_update(v, cfg.get(c, {}))
                print(f'{v=}')
                set_mode(c, v.pop('mode'), update=False)
                set_inst(
                    c, v.pop('inst'), 
                    program=v.pop('program_inst', None), 
                    update=False)
                set_mute(c, v.pop('mute'), update=False)
                set_cc(c, v.pop('cc'), update=False)
                config[c].update(v)

        if update:
            update_config()

    ### action_* runs on key/button press;
    ### invokes cycler / picker logic and schedules set_*

    # this is pretty awful
    # need a better way to reconcile iipyper and textual here
    def action_mode(c):
        if c not in config: return
        # TODO: mode picker
        if config[c]['mode'] == 'auto':
            mode = 'input'
        elif config[c]['mode'] == 'input' and config[c]['source']!=c:
            # TODO: source picker for follow
            mode = 'follow'
        else:
            mode = 'auto'
        action_queue.append(ft.partial(set_mode, c, mode))

    def action_inst(c):
        print(f'inst channel {c}')
        tui.push_screen(InstrumentGroupSelect(c))

    def action_mute(c):
        if i not in config: return
        # set_mute(c, not config[c].get('mute', False))
        action_queue.append(ft.partial(
            set_mute, c, not config[c].get('mute', False)))

    def action_preset(p):
        if p >= len(presets): 
            return
        action_queue.append(ft.partial(set_preset, p))

    ### set actions which have an index argument
    ### TODO move this logic into @tui.set_action

    for i in range(1,17):
        setattr(tui, f'action_mode_{i}', ft.partial(action_mode, i))
        setattr(tui, f'action_inst_{i}', ft.partial(action_inst, i))
        setattr(tui, f'action_mute_{i}', ft.partial(action_mute, i))

    for i in range(NotoPresets.n_presets):
        setattr(tui, f'action_preset_{i}', ft.partial(action_preset, i))

    ### additional key/button actions

    @tui.set_action
    def mute():
        action_queue.append(noto_mute)

    @tui.set_action
    def sustain():
        action_queue.append(ft.partial(noto_mute, sustain=True))
    
    @tui.set_action
    def reset():
        action_queue.append(noto_reset)
    
    @tui.set_action
    def query():
        action_queue.append(auto_query)

    @tui.set_action
    def stop():
        action_queue.append(noto_stop)

    @tui.set_action
    def preset1():
        action_queue.append(ft.partial(set_preset, 0))
    @tui.set_action
    def preset2():
        action_queue.append(ft.partial(set_preset, 1))
    @tui.set_action
    def preset3():
        action_queue.append(ft.partial(set_preset, 2))
    @tui.set_action
    def preset4():
        action_queue.append(ft.partial(set_preset, 3))
    @tui.set_action
    def preset5():
        action_queue.append(ft.partial(set_preset, 4))
    @tui.set_action
    def preset6():
        action_queue.append(ft.partial(set_preset, 5))
    @tui.set_action
    def preset7():
        action_queue.append(ft.partial(set_preset, 6))
    @tui.set_action
    def preset8():
        action_queue.append(ft.partial(set_preset, 7))
    @tui.set_action
    def preset9():
        action_queue.append(ft.partial(set_preset, 8))
    @tui.set_action
    def preset10():
        action_queue.append(ft.partial(set_preset, 9))

    ### TUI classes which close over variables defined in main

    class Instrument(Button):
        """button which picks an instrument"""
        def __init__(self, c, i):
            super().__init__(inst_label(i))
            self.channel = c
            self.inst = i
        def on_button_pressed(self, event: Button.Pressed):
            self.app.pop_screen()
            self.app.pop_screen()
            # set_inst(self.channel, self.inst)
            action_queue.append(ft.partial(
                set_inst, self.channel, self.inst))

    class InstrumentGroup(Button):
        """button which picks an instrument group"""
        def __init__(self, text, c, g):
            super().__init__(text)
            self.channel = c
            self.group = g
        def on_button_pressed(self, event: Button.Pressed):
            # show inst buttons
            tui.push_screen(InstrumentSelect(self.channel, self.group))

    class InstrumentSelect(ModalScreen):
        """Screen with instruments"""
        def __init__(self, c, g):
            super().__init__()
            self.channel = c
            self.group = g

        def compose(self):
            yield Grid(
                *(
                    Instrument(self.channel, i)
                    for i in gm_groups[self.group][1]
                ), id="dialog",
            )

    class InstrumentGroupSelect(ModalScreen):
        """Screen with instrument groups"""
        # TODO: add other features to this screen -- transpose, range, etc?
        def __init__(self, c):
            super().__init__()
            self.channel = c

        def compose(self):
            yield Grid(
                *(
                    InstrumentGroup(s, self.channel, g)
                    for g,(s,_) in enumerate(gm_groups)
                ), id="dialog",
            )

    initial_preset = 0
    if isinstance(preset, str):
        for n,p in enumerate(presets):
            if preset == p['name']:
                initial_preset = n
                break

    config_cli = config or {}
    config = {i:default_config_channel(i) for i in range(1,17)}
    set_preset(initial_preset, update=False)
    if config_ingest is not None:
        set_config(config_ingest, overlay=False, update=False)
    set_config(config_cli, overlay=True, update=False)

    prediction_displayed = [None]
    @repeat(lock=True, err_file=tui)
    # @profile(print=print, enable=profiler)
    def _():
        """Loop, process enqueued actions and check if predicted next event happens"""
        # print(f'repeat {time.time()}')
        for _ in range(8): # process multiple actions per tick
            if len(action_queue):
                # print(action_queue)
                action_queue.pop(0)()
        # TODO: immediate query if an input has just happened

        # if there is no predicted event,
        # or it's not for an auto voice,
        # sample one:
        if pending.gate and not pending.stopped and not pending.is_auto():
            # with profile('auto_query', print=print, enable=profiler):
            auto_query()

        # if unmuted, predicted event is auto, and its time has passed,
        # realize it
        if (
            not testing and
            pending.gate and
            pending.is_auto() and
            pending.time_until() <= 0
            # pending.time_until() <= estimated_feed_latency
            ):
            # with profile('auto_event', print=print, enable=profiler):
            auto_event()


            # query for new prediction
            if not (pending.stopped or debug_query):
                # with profile('auto_query', print=print, enable=profiler):
                auto_query(immediate=True)
        else:
            # otherwise there is a pause, update the UI with next prediction
            if pending.event != prediction_displayed[0]:
                tui.defer(prediction=pending.event)
                prediction_displayed[0] = pending.event
            maybe_punch_out()

        # TODO note sure if this is needed anymore
        # adaptive time resolution here -- yield to other threads when 
        # next event is not expected, but time precisely when next event
        # is imminent
        wait = 10e-3
        if len(action_queue):
            wait = 0
        elif pending.is_auto():
            wait = pending.time_until()
        # print(f'{wait=}')
        r = max(0, min(10e-3, wait))
        return r

    @tui.on
    def mount(): 
        update_config()
        update_presets()
        print('MIDI handling:')
        print(midi.get_docs())
        if osc_port is not None:
            print('OSC handling:')
            print(osc.get_docs())
        print(tui_doc)
        print('For more detailed documentation, see:')
        print('https://intelligent-instruments-lab.github.io/notochord/reference/notochord/app/homunculus/')
        print('or run `notochord homunculus --help`')
        print('to exit, use CTRL+C')

        action_queue.append(noto_reset)

        if initial_stop:
            action_queue.append(noto_stop)
        else:
            action_queue.append(ft.partial(
                auto_query, predict_input=False, predict_follow=False))

    if use_tui:
        tui.run()
    else:
        if initial_stop:
            noto_stop()
        else:
            auto_query(predict_input=False, predict_follow=False)
        

### def TUI components ###
class NotoLog(Log):
    value = reactive('')
    def watch_value(self, time: float) -> None:
        self.write_line(self.value)

class NotoPrediction(Static):
    value = reactive(None)
    def watch_value(self, time: float) -> None:
        evt = self.value
        if evt is None:
            s = ''
        else:
            s = f"\tinstrument: {evt['inst']:3d}    pitch: {evt['pitch']:3d}    time: {int(evt['time']*1000):4d} ms    velocity:{int(evt['vel']):3d}     end: {evt.get('end', 0):.5f}"
        self.update(Panel(s, title='prediction'))

class Mixer(Static):
    def compose(self):
        for i in range(1,17):
            yield MixerButtons(i, id=f"mixer_{i}")

def preset_id(i):
    return f"preset_{i}"
def mode_id(i):
    return f"mode_{i}"
def inst_id(i):
    return f"inst_{i}"
def mute_id(i):
    return f"mute_{i}"
class MixerButtons(Static):
    def __init__(self, idx, **kw):
        self.idx = idx
        super().__init__(**kw)

    def compose(self):
        yield Button(
            f"{self.idx:02d}", 
            id=mode_id(self.idx),
            classes="cmode"
            )
        yield Button(
            f"--- \n-----\n-----", 
            id=inst_id(self.idx),
            classes="cinst"
            )
        yield Button(
            f"MUTE", 
            id=mute_id(self.idx),
            classes="cmute"
            )
        
    def on_mount(self) -> None:
        lines = [
            "cycle channel mode:",
            f"-->{self.idx:02d}   input from MIDI channel {self.idx}"
        ]
        if self.idx > 0: lines.append(
            f"{self.idx-1:02d}->{self.idx:02d}  follow channel {self.idx-1}")
        lines.append(
            f"{self.idx:02d}      notochord plays autonomously"
        )
        self.query_one("#"+mode_id(self.idx)).tooltip = '\n'.join(lines)
        self.query_one("#"+inst_id(self.idx)).tooltip = "open instrument picker"
        self.query_one("#"+mute_id(self.idx)).tooltip = f"mute channel {self.idx}"
        
class NotoPresets(Static):
    n_presets = 10
    def compose(self):
        for i in range(NotoPresets.n_presets):
            yield Button('---', id=preset_id(i))

    def on_mount(self) -> None:
        for i in range(NotoPresets.n_presets):
            self.query_one("#"+preset_id(i)).tooltip = f"load preset {i+1}"

class NotoControl(Static):
    def compose(self):
        # yield NotoToggle()
        # yield Checkbox("Mute", id="mute")
        yield Button("Mute", id="mute", variant="error")
        yield Button("Sustain", id="sustain", variant="primary")
        yield Button("Query", id="query")
        yield Button("Stop", id="stop")
        yield Button("Reset", id="reset", variant="warning")

    def on_mount(self) -> None:
        self.query_one("#mute").tooltip = "master mute notochord"
        self.query_one("#sustain").tooltip = "master sustain -- prevent any NoteOffs from notochord"
        self.query_one("#query").tooltip = "manually query a new event"
        self.query_one("#reset").tooltip = "reset notochord"

class NotoTUI(TUI):
    CSS_PATH = 'homunculus.css'

    BINDINGS = [
        ("m", "mute", "Mute"),
        ("s", "sustain", "Sustain"),
        ("q", "query", "Re-query"),
        ("x", "stop", "Stop"),
        ("r", "reset", "Reset"),
        ("1", "preset1", ""),
        ("2", "preset2", ""),
        ("3", "preset3", ""),
        ("4", "preset4", ""),
        ("5", "preset5", ""),
        ("6", "preset6", ""),
        ("7", "preset7", ""),
        ("8", "preset8", ""),
        ("9", "preset9", ""),
        ("0", "preset10", ""),
        ]
    
    def compose(self):
        """Create child widgets for the app."""
        yield Header()
        yield self.std_log
        yield NotoLog(id='note')
        yield NotoPrediction(id='prediction')
        yield Mixer()
        yield NotoPresets()
        yield NotoControl()
        yield Footer()

    def on_mount(self) -> None:
        self.query_one(NotoPrediction).tooltip = "displays the next predicted event"
        # print(self.screen)

    def set_preset(self, idx, name):
        if idx < NotoPresets.n_presets:
            node = self.query_one('#'+preset_id(idx))
            node.label = str(idx) if name is None else name
        else:
            self.write(f'warning: more than {NotoPresets.n_presets} presets\n')
        # node.label = name

    def set_channel(self, chan, cfg):
        # print(f'set_channel {cfg}')
        inst_node = self.query_one('#'+inst_id(chan))
        mode_node = self.query_one('#'+mode_id(chan))
        mute_node = self.query_one('#'+mute_id(chan))

        if cfg is None:
            inst_node.variant = 'default'
            mute_node.variant = 'default'
            mode_node.variant = 'default'
            return
        
        mode = cfg['mode']
        mute = cfg['mute']
        inst = cfg['inst']

        inst_node.label = inst_label(inst)

        if mode=='auto':
            mode_node.label = f"{chan:02d}"
        elif mode=='input':
            mode_node.label = f"-->{chan:02d}"
        elif mode=='follow':
            mode_node.label = f"{cfg['source']:02d}->{chan:02d}"

        if mute:
            mode_node.variant = 'default'
            inst_node.variant = 'default'
            mute_node.label = 'UNMUTE'
            mute_node.variant = 'error'
        else:
            mode_node.variant = 'primary'
            inst_node.variant = 'warning'
            mute_node.label = 'MUTE'
            mute_node.variant = 'default'

gm_names = [
    '_SEQ_\nSTART',
    'GRAND\nPIANO', 'BRGHT\nPIANO', 'EGRND\nPIANO', 'HONKY\n-TONK', 
    'RHODE\nPIANO', 'FM   \nPIANO', 'HRPSI\nCHORD', 'CLAV \n INET',
    'CEL  \n ESTA', 'GLOCN\nSPIEL', 'MUSIC\n BOX ', 'VIBRA\nPHONE', 
    'MAR  \n IMBA', 'XYLO \nPHONE', 'TUBLR\n BELL', 'DULCI\n  MER',
    'DRAWB\nORGAN', 'PERC \nORGAN', 'ROCK \nORGAN', 'CHRCH\nORGAN', 
    'REED \nORGAN', 'ACCOR\n DION', 'HARMO\n NICA', 'BANDO\n NEON',  
    'A-GTR\nNYLON', 'A-GTR\nSTEEL', 'E-GTR\nJAZZ ', 'E-GTR\nCLEAN', 
    'E-GTR\nMUTED', 'E-GTR\nDRIVE', 'E-GTR\nDIST ', 'E-GTR\nHRMNC',  
    'A-GTR\n BASS', 'EBASS\nFINGR', 'EBASS\n PICK', 'EBASS\nFRTLS', 
    'SLAP \nBASS1', 'SLAP \nBASS2', 'SYNTH\nBASS1', 'SYNTH\nBASS2',  
    'STRNG\nVIOLN', 'STRNG\nVIOLA', 'STRNG\nCELLO', 'STRNG\nCBASS', 
    'STRNG\nTREMO', 'STRNG\nPIZZC', 'ORCH \n HARP', 'TIMP \n  ANI',  
    'STRNG\nENSB1', 'STRNG\nENSB2', 'SYNTH\nSTRG1', 'SYNTH\nSTRG2', 
    'CHOIR\n AAH ', 'CHOIR\n OOH ', 'SYNTH\nVOICE', 'ORCH \n  HIT',  
    'TRUM \n  PET', 'TROM \n BONE', 'TUBA \n     ', 'MUTED\nTRMPT', 
    'FRNCH\nHORN ', 'BRASS\nSECTN', 'SYNTH\nBRSS1', 'SYNTH\nBRSS2',  
    'SPRNO\n SAX ', 'ALTO \n  SAX', 'TENOR\n  SAX', 'BARI \n  SAX', 
    'OBOE \n     ', 'ENGLS\nHORN ', 'BASS \n  OON', 'CLARI\n  NET',  
    'PICC \n  OLO', 'FLUTE\n     ', 'RECO \n RDER', ' PAN \nFLUTE',
    'BLOWN\nBOTTL', 'SHAKU\nHACHI', 'WHIS \n  TLE', 'OCA  \n RINA',  
    'LEAD1\nSQUAR', 'LEAD2\n SAW ', 'LEAD3\n TRI ', 'LEAD4\nCHIFF',
    'LEAD5\nCHRNG', 'LEAD6\nVOICE', 'LEAD7\nFIFTH', 'LEAD8\nSYNTH',  
    'PAD 1\nNWAGE', 'PAD 2\n WARM', 'PAD 3\n POLY', 'PAD 4\nCHOIR',
    'PAD 5\nGLASS', 'PAD 6\nMETAL', 'PAD 7\n HALO', 'PAD 8\nSWEEP',
    'FX  1\n RAIN', 'FX  2\nSDTRK', 'FX  3\nCRYST', 'FX  4\nATMOS',
    'FX  5\nBRGHT', 'FX  6\nGOBLN', 'FX  7\nECHOS', 'FX  8\nSCIFI',
    'SITAR\n     ', 'BANJO\n     ', 'SHAM \n ISEN', 'KOTO \n     ',
    'KAL  \n IMBA', 'BAG  \n PIPE', 'FID  \n  DLE', 'SHA  \n  NAI',  
    'TINKL\nBELL ', 'AGO  \n   GÔ', 'STEEL\nDRUM ', 'WOOD \nBLOCK', 
    'TAIKO\nDRUM ', 'MELO \n  TOM', 'SYNTH\nDRUM ', ' REV \nCYMBL',  
    'GTR  \n FRET', 'BRE  \n  ATH', ' SEA \nSHORE', 'BIRD \nTWEET',
    'TELE \nPHONE', 'HELI \nCOPTR', 'APP  \nLAUSE', 'GUN  \n SHOT',  
] + (
    ['STD  \n  KIT']*8 + 
    ['ROOM \n  KIT']*8 + 
    ['ROCK \n  KIT']*8 + 
    ['ELCTR\n  KIT']*8 + 
    ['JAZZ \n  KIT']*8 + 
    ['BRUSH\n  KIT']*8 + 
    ['ORCHS\n  KIT']*8 + 
    ['SFX  \n  KIT']*8 + 
    ['DRUM \n KIT?']*64 
) + ['ANON \n MEL ']*32 + ['ANON \nDRUMS']*32
gm_groups = [
    ('PIANO\n     ', range(1,9)),
    ('CHROM\nPERC ', range(9,17)),
    ('ORGAN\n     ', range(17,25)),
    ('GUI  \n TARS', range(25,33)),
    ('BASS \n GTRS', range(33,41)),
    ('STRIN\n   GS', range(41,49)),
    ('ENSEM\n BLES', range(49,57)),
    ('BRASS\n     ', range(57,65)),
    ('REEDS\n     ', range(65,73)),
    ('PIPES\n     ', range(73,81)),
    ('SYNTH\nLEADS', range(81,89)),
    ('SYNTH\nPADS ', range(89,97)),
    ('SYNTH\nFX   ', range(97,105)),
    ('MISC \n  MEL', range(105,113)),
    ('MISC \n PERC', range(113,121)),
    ('SOUND\nFX   ', range(121,129)),
    ('DRUM \n KITS', [128+i for i in (1,9,17,25,33,41,49,57)]),
    # (' STD \nDRUMS', range(129,137)),
    # ('ROOM \nDRUMS', range(137,145)),
    # ('ROCK \nDRUMS', range(145,153)),
    # ('ELCTR\nDRUMS', range(153,161)),
    # ('JAZZ \nDRUMS', range(161,169)),
    # ('BRUSH\nDRUMS', range(169,177)),
    # ('ORCH \nDRUMS', range(177,185)),
    # (' SFX \nDRUMS', range(185,193)),
    ('ANON\n  MEL', range(257,273)),
    ('ANON\n DRUM', range(289,305)),
]
def inst_label(i):
    if i is None:
        return f"--- \n-----\n-----"
    return f'{i:03d} \n{gm_names[i]}'
### end def TUI components###

if __name__=='__main__':
    run(main)



"""
Authors:
  Victor Shepardson
  Intelligent Instruments Lab 2024
"""

from notochord import Notochord

def main(
        midi_file, 
        checkpoint="notochord-latest.ckpt",
        start_tokens=True
        ):
    """
    Run the contents of a MIDI file through a Notochord model and cache the results.

    Args:
        midi_file: path to a MIDI file to prompt with.
        checkpoint: path to Notochord model.
        start_tokens: whether to include send start-of-sequence tokens in prompt
    """
    print(f'{midi_file=} {checkpoint=}')
    noto = Notochord.from_checkpoint(checkpoint)
    noto.eval()
    noto.reset(start=start_tokens)
    _, inst_map = noto.prompt(midi_file)
    print('MIDI channel to instrument:', inst_map)


"""
Notochord MIDI co-improviser server.
Notochord plays different instruments along with the player.

Authors:
  Victor Shepardson
  Intelligent Instruments Lab 2023
"""

from typing import Optional, Dict
from numbers import Number
import random

from notochord import Notochord, MIDIConfig, NotoPerformance
from iipyper import OSC, MIDI, run, Stopwatch, repeat, cleanup, TUI, profile, lock

from rich.panel import Panel
from rich.pretty import Pretty
from textual.reactive import reactive
from textual.widgets import Header, Footer, Static, Button, Log

### def TUI components ###
class NotoLog(Log):
    value = reactive('')
    def watch_value(self, time: float) -> None:
        self.write(self.value)

class NotoPrediction(Static):
    value = reactive(None)
    def watch_value(self, time: float) -> None:
        evt = self.value
        if evt is None:
            s = ''
        else:
            s = f"\tinstrument: {evt['inst']:3d}    pitch: {evt['pitch']:3d}    time: {int(evt['time']*1000):4d} ms    velocity:{int(evt['vel']):3d}     end: {evt['end']:4f}"
        self.update(Panel(s, title='prediction'))

class NotoControl(Static):
    def compose(self):
        yield Button("Mute", id="mute", variant="error")
        yield Button("Query", id="query")
        yield Button("Reset", id="reset", variant='warning')

class NotoTUI(TUI):
    CSS_PATH = 'improviser.css'

    BINDINGS = [
        ("m", "mute", "Mute Notochord"),
        ("q", "query", "Re-query Notochord"),
        ("r", "reset", "Reset Notochord")]

    def compose(self):
        """Create child widgets for the app."""
        yield Header()
        yield self.std_log
        yield NotoLog(id='note')
        yield NotoPrediction(id='prediction')
        yield NotoControl()
        yield Footer()
### end def TUI components###

def main(
    checkpoint="txala-latest.ckpt", # Notochord checkpoint
    player_config:Dict[int,int]=None, # map MIDI channel : GM instrument
    noto_config:Dict[int,int]=None, # map MIDI channel : GM instrument
    pitch_set=(41,43,45), # MIDI pitches used
    vel_range=(80,120),
    # scale_vel=False,

    initial_mute=False, # start with Notochord muted
    initial_query=False, # let Notochord start playing immediately

    midi_in:Optional[str]=None, # MIDI port for player input
    midi_out:Optional[str]=None, # MIDI port for Notochord output
    thru=False, # copy player input to output
    dump_midi=False, # print all incoming MIDI

    input_latency=0.005,
    rhythm_temp=0.9,
    timing_temp=0.05,
    steer_rate=0.55,

    auto_reset=True,

    start_after=2, # don't sample notochord until this many total events
    max_run=5,
    min_run=1,
    
    min_time=0,
    max_time=None, # max time between events
    nominal_time=False, #feed Notochord with nominal dt instead of actual
    backoff_time=50e-3, #time to wait when a predicted player event doesn't happen

    osc_port=None, # if supplied, listen for OSC to set controls on this port
    osc_host='', # leave this as empty string to get all traffic on the port

    use_tui=True, # run textual UI
    predict_player=True, # forecasted next events can be for player (preserves model distribution, but can lead to Notochord deciding not to play)
    auto_query=True, # query notochord whenever it is unmuted and there is no pending event. generally should be True except for testing purposes.
    testing=False,
    verbose=0
    ):
    """
    Args:
        checkpoint: path to notochord model checkpoint.

        player_config: mapping from MIDI channels to MIDI instruments controlled
            by the player.
        noto_config: mapping from MIDI channels to MIDI instruments controlled
            by notochord. Both indexed from 1.
            instruments should be different from the player instruments.
            channels should be different unless different ports are used.
            MIDI channels and General MIDI instruments are indexed from 1.

        pitch_set: collection of MIDI pitches for the txalaparta boards
        vel_range: range of velocities used

        initial_mute: start Notochord muted so it won't play with input.
        initial_query: query Notochord immediately so it plays even without input.

        midi_in: MIDI ports for player input. 
            default is to use all input ports.
            can be comma-separated list of ports.
        midi_out: MIDI ports for Notochord output. 
            default is to use only virtual 'From iipyper' port.
            can be comma-separated list of ports.
        thru: if True, copy incoming MIDI to output ports.
            only makes sense if input and output ports are different.
        dump_midi: if True, print all incoming MIDI for debugging purposes

        min_time: minimum time in seconds between predicted events.
            default is 0.
        max_time: maximum time in seconds between predicted events.
            default is the Notochord model's maximum (usually 10 seconds).
        nominal_time: if True, feed Notochord with its own predicted times
            instead of the actual elapsed time.
        backoff_time: time in seconds to wait before querying,
            if a predicted player event doesn't occur

        input_latency: estimated input latency in seconds
        rhythm_temp: temperature for sampling time mixture component,
        timing_temp: temperature for sampling time component distribution
        steer_rate: 0.5 is unbiased; values >0.5 tend faster, <0.5 tend slower

        auto_reset: reset the notochord model when it predicts end of sequence
        start_after: don't sample notochord until this many total events
        max_run: player of the next hit must change after this many 
            consecutive hits. can also be a dict with different values 
            for p1 and p2
        min_run: player of the next hit cannot change before this many 
            consecutive hits. can also be a dict with different values 
            for p1 and p2

        osc_port: optional. if supplied, listen for OSC to set controls
        osc_host: hostname or IP of OSC sender.
            leave this as empty string to get all traffic on the port

        use_tui: run textual UI.
        predict_player: forecasted next events can be for player.
            generally should be True;
            instead use balance_sample to force Notochord to play.
        auto_query: query notochord whenever it is unmuted and there is no pending event. generally should be True unless debugging.
    """
    if osc_port is not None:
        osc = OSC(osc_host, osc_port)
    midi = MIDI(midi_in, midi_out)

    ### Textual UI
    tui = NotoTUI()
    print = tui.print
    ###

    if isinstance(max_run, Number):
        max_run = {'p1':max_run, 'p2':max_run}
    if isinstance(min_run, Number):
        min_run = {'p1':min_run, 'p2':min_run}

    max_run[None] = 99
    min_run[None] = 0

    if isinstance(max_time, Number) or max_time is None:
        max_time = {'p1':max_time, 'p2':max_time}
    if isinstance(min_time, Number) or min_time is None:
        min_time = {'p1':min_time, 'p2':min_time}
    max_time[None] = None
    min_time[None] = 0

    # default channel:instrument mappings
    if player_config is None:
        # player_config = {1:265,2:266}
        player_config = {1:290,2:291}
    if noto_config is None:
        # noto_config = {3:267,4:268}
        noto_config = {3:292,4:293}

    state = {
        'run_length': 0,
        'total_count': 0,
        'last_side': None
    }

    # assuming 2 player 2 notochord or else 4 notochord
    if len(player_config)==2:
        p1_insts = set(player_config.values())
        p2_insts = set(noto_config.values())
    elif len(noto_config)==4:
        # if all notochord, ass
        insts = sorted(noto_config.values())
        p1_insts = set(insts[:2])
        p2_insts = set(insts[2:])
    else:
        raise ValueError("""
            invalid config.
            there should be two player and two notochord hands, 
            or four notochord hands.
        """)

    # convert 1-indexed MIDI channels to 0-indexed here
    player_map = MIDIConfig({k-1:v for k,v in player_config.items()})
    noto_map = MIDIConfig({k-1:v for k,v in noto_config.items()})
    
    if len(player_map.insts & noto_map.insts):
        print("WARNING: Notochord and Player instruments shouldn't overlap")
        print('setting to an anonymous instrument')
        # TODO: set to anon insts without changing mel/drum
        # respecting anon insts selected for player
        raise NotImplementedError
    # TODO:
    # check for repeated insts/channels

    def warn_inst(i):
        if i > 128:
            if i < 257:
                print(f"WARNING: drum instrument {i} selected, be sure to select a drum bank in your synthesizer")
            else:
                print(f"WARNING: instrument {i} is not General MIDI")
    
    # load notochord model
    try:
        noto = Notochord.from_checkpoint(checkpoint)
        noto.eval()
        noto.reset()
    except Exception:
        print("""error loading notochord model""")
        raise

    # main stopwatch to track time difference between MIDI events
    stopwatch = Stopwatch()

    # simple class to hold pending event prediction
    class Prediction:
        def __init__(self):
            self.event = None
            self.gate = not initial_mute
    pending = Prediction()

    # query parameters controlled via MIDI / OSC
    controls = {}
    if steer_rate is not None:
        controls['steer_rate'] = steer_rate

    # tracks held notes, recently played instruments, etc
    # history = NotoPerformance()

    def display_event(tag, memo, inst, pitch, vel, channel, **kw):
        """print an event to the terminal"""
        if tag is None:
            return
        s = f'{tag}:\t {inst=:4d}    {pitch=:4d}    {vel=:4d}    {channel=:3d}'
        if memo is not None:
            s += f'    ({memo})'
        s += '\n'
        tui.defer(note=s)

    def play_event(event, channel, feed=True, send=True, tag=None, memo=None):
        """realize an event as MIDI, terminal display, and Notochord update"""
        # normalize values
        vel = event['vel'] = round(event['vel'])
        dt = stopwatch.punch(latency=input_latency if tag=='PLAYER' else 0)

        if 'time' not in event or not nominal_time:
            event['time'] = dt

        # side = 'player' if event['inst'] in player_config.values() else 'noto'
        side = 'p1' if event['inst'] in p1_insts else 'p2'
        if side==state['last_side']:
            state['run_length'] += 1
        else:
            state['run_length'] = 1

        state['total_count'] += 1

        state['last_side'] = side
        if verbose > 0:
            print(f'{state["run_length"]=}')

        # send out as MIDI
        if send:
            midi.send(
                'note_on' if vel > 0 else 'note_off', 
                note=event['pitch'], velocity=vel, channel=channel)

        # print
        display_event(tag, memo=memo, channel=channel, **event)

        # feed to Notochord
        if feed:
            noto.feed(**event)

    # @lock
    def noto_reset(query=True):
        """reset Notochord and end all of its held notes"""
        print('RESET')

        # cancel pending predictions
        pending.event = None
        tui.defer(prediction=pending.event)
        
        # reset stopwatch
        stopwatch.punch()
        # reset notochord state
        noto.reset()

        state['total_count'] = 0
        # reset history
        # history.push()
        # query the fresh notochord for a new prediction
        if pending.gate and query:
            noto_query()

    # @lock
    def noto_mute():
        pending.gate = not pending.gate
        print('UNMUTE' if pending.gate else 'MUTE')
        # if unmuting, we're done
        if pending.gate:
            return
        # cancel pending predictions
        pending.event = None
        tui(prediction=pending.event)

    # query Notochord for a new next event
    # @lock
    def noto_query(delay=0, predict_noto=True):

        all_insts = noto_map.insts 
        if predict_player:
            all_insts |= player_map.insts

        steer_time = 1-controls.get('steer_rate', 0.5)
        tqt = (max(0,steer_time-0.5), min(1, steer_time+0.5))

        if verbose > 1:
            print(tqt)

        last_side = state['last_side']
        # if using nominal time,
        # *subtract* estimated feed latency to min_time; (TODO: really should
        #   set no min time when querying, use stopwatch when re-querying...)
        # if using actual time, *add* estimated query latency
        time_offset = -5e-3 if nominal_time else 10e-3
        min_t = max(min_time[last_side], stopwatch.read()+time_offset+delay)


        # print(state)
        if state['total_count'] < start_after or not predict_noto:
            insts = player_map.insts
        else:
            insts = all_insts
            last_side_insts = p1_insts if last_side=='p1' else p2_insts
            if state['run_length'] >= max_run[last_side]:
                # if state['last_side']=='noto':
                    # s = set(noto_config.values()) 
                # else:
                    # s = set(player_config.values()) 
                insts = insts - last_side_insts
            elif state['run_length'] < min_run[last_side]:
                insts = last_side_insts


        if len(insts)==0:
            insts = all_insts

        if verbose > 1:
            print(f'{insts=}')


        if len(insts) > 2:
            # in this case *only* time is messed with,
            # so if we sample time first,
            # the rest can be properly conditioned on it
            query_method = noto.query_tipv_onsets
        else:
            # in this case, instrument and time have both been constrained,
            # and we can't sample the true joint distribution,
            # but we sample instrument first
            # under the assumption that the instrument constraint is
            # 'stronger' than the time constraint
            query_method = noto.query_itpv_onsets

        max_t = max_time[last_side]
        max_t = None if max_t is None else max(max_t, min_t+0.03)

        if verbose > 1:
            print(f'{min_time=}, {max_t=}')

        pending.event = query_method(
            include_pitch=pitch_set,
            include_inst=list(insts),
            min_time=min_t, max_time=max_t,
            truncate_quantile_time=tqt,
            min_vel=vel_range[0], max_vel=vel_range[1],
            rhythm_temp=rhythm_temp,
            timing_temp=timing_temp,
        )
        # display the predicted event
        tui.defer(prediction=pending.event)

    #### MIDI handling

    # print all incoming MIDI for debugging
    if dump_midi:
        @midi.handle
        def _(msg):
            print(msg)

    # @midi.handle(type='program_change')
    # def _(msg):
    #     """Program change events set instruments"""
    #     if msg.channel in player_map:
    #         player_map[msg.channel] = msg.program
    #     if msg.channel in noto_map:
    #         noto_map[msg.channel] = msg.program

    # @midi.handle(type='pitchwheel')
    # def _(msg):
    #     controls['steer_pitch'] = (msg.pitch+8192)/16384
        # print(controls)

    # very basic CC handling for controls
    @midi.handle(type='control_change')
    def _(msg):
        """CC messages on any channel"""

        # if msg.control==1:
        #     controls['steer_pitch'] = msg.value/127
        #     print(f"{controls['steer_pitch']=}")
        # if msg.control==2:
        #     controls['steer_density'] = msg.value/127
        #     print(f"{controls['steer_density']=}")
        if msg.control==3:
            controls['steer_rate'] = msg.value/127
            if verbose > 0:
                print(f"{controls['steer_rate']=}")

        if msg.control==4:
            noto_reset()
        if msg.control==5:
            noto_query()
        if msg.control==6:
            noto_mute()

    # very basic OSC handling for controls
    if osc_port is not None:
        @osc.args('/notochord/improviser/*')
        def _(route, *a):
            print('OSC:', route, *a)
            ctrl = route.split['/'][3]
            if ctrl=='reset':
                noto_reset()
            elif ctrl=='query':
                noto_query()
            elif ctrl=='mute':
                noto_mute()
            else:
                assert len(a)==0
                arg = a[0]
                assert isinstance(arg, Number)
                controls[ctrl] = arg
                print(controls)

    # @midi.handle(type=('note_on', 'note_off'))
    @midi.handle(type='note_on')
    def _(msg):
        """MIDI NoteOn events from the player"""
        # if thru and msg.channel not in noto_map.channels:
            # midi.send(msg)

        if msg.channel not in player_map.channels:
            return
        
        inst = player_map[msg.channel]
        pitch = msg.note
        vel = msg.velocity

        if vel > 0:
            # feed event to Notochord
            # with profile('feed', print=print):
            play_event(
                {'inst':inst, 'pitch':pitch, 'vel':vel}, 
                channel=msg.channel, send=thru, tag='PLAYER (p1)')

            # query for new prediction
            noto_query()

        # send a MIDI reply for latency testing purposes:
        # if testing: midi.cc(control=3, value=msg.note, channel=15)

    def noto_event():
        # notochord event happens:
        event = pending.event
        inst, pitch, vel = event['inst'], event['pitch'], round(event['vel'])

        # note on which is already playing or note off which is not
        # if (vel>0) == ((inst, pitch) in history.note_pairs): 
        #     print(f're-query for invalid {vel=}, {inst=}, {pitch=}')
        #     noto_query()
        #     return
        
        chan = noto_map.inv(inst)
        p = 'p1' if inst in p1_insts else 'p2'
        play_event(event, channel=chan, tag=f'NOTO ({p})')

    @repeat(1e-3, lock=True)
    def _():
        """Loop, checking if predicted next event happens"""
        # check if current prediction has passed
        if (
            not testing and
            pending.gate and
            pending.event is not None and
            stopwatch.read() > pending.event['time']
            ):
            e = pending.event
            if e['time'] >= noto.time_dist.hi.item():
                if auto_reset:
                    noto_reset(query=False)
            # if so, check if it is a notochord-controlled instrument
            delay = 0
            predict_noto = True
            if e['inst'] in noto_map.insts:
                # prediction happens
                noto_event()
            else:
                # if last event was noto, still predict player only
                if state['last_side']=='noto':
                    predict_noto = False
                else:
                    # otherwise, hesitate before possibly predicting noto turn
                    delay = backoff_time
            # query for new prediction
            # print(pending.event)
            pending.event = None
            if ('end' in e and random.random() < e['end']):
                print('END')
                if auto_reset:
                    noto_reset(query=False)
            elif auto_query:
                noto_query(delay=delay, predict_noto=predict_noto)

    @cleanup
    def _():
        """"""
        pass
        # print(f'cleanup: {notes=}')
        # for (chan,inst,pitch) in history.note_triples:
        # # for (inst,pitch) in notes:
        #     if inst in noto_map.insts:
        #         midi.note_on(note=pitch, velocity=0, channel=chan)

    @tui.set_action
    def mute():
        noto_mute()
    
    @tui.set_action
    def reset():
        noto_reset()
    
    @tui.set_action
    def query():
        noto_query()

    if initial_query:
        noto_query()

    if use_tui:
        tui.run()


if __name__=='__main__':
    run(main)



import sys, subprocess

import fire
from iipyper import run

from notochord import Notochord
from notochord.app import *

def help():
    print("""
    available subcommands:
        server: run the Notochord API OSC server
        homunculus: run the Notochord homunculus TUI
        harmonizer: run the Notochord harmonizer TUI
        improviser: run the Notochord improviser TUI
        txalaparta: run the txalaparta app
        morse: run the morse code app
        prompt: run a MIDI file through a notochord model and cache hidden states
        train: train a Notochord model (GPU recommended)
        files: show the location of Notochord models and config files on disk
    """)

def _main():
    # print(sys.argv)
    try:
        if sys.argv[1] == 'server':
            sys.argv = sys.argv[1:]
            run(server)
        if sys.argv[1] == 'homunculus':
            sys.argv = sys.argv[1:]
            run(homunculus) 
        if sys.argv[1] == 'harmonizer':
            sys.argv = sys.argv[1:]
            run(harmonizer)
        if sys.argv[1] == 'improviser':
            sys.argv = sys.argv[1:]
            run(improviser)
        if sys.argv[1] == 'txalaparta':
            sys.argv = sys.argv[1:]
            run(txalaparta)
        if sys.argv[1] == 'morse':
            sys.argv = sys.argv[1:]
            run(morse)
        if sys.argv[1] == 'prompt':
            fire.Fire(prompt, sys.argv[2:])
            return
        if sys.argv[1] == 'train':
            from notochord.train import Resumable
            fire.Fire(Resumable, sys.argv[1:])
        if sys.argv[1] == 'files':
            d = Notochord.user_data_dir()
            print(d)
            if sys.platform=='darwin':
                subprocess.run(('open', d))
            elif sys.platform=='win32':
                os.startfile(d)
        else:
            help()
    except IndexError:
        help()

if __name__=='__main__':
    _main()



"""
Notochord MIDI harmonizer server.
Each note from the player produces a harmonizing note from Notochord.

Authors:
  Victor Shepardson
  Intelligent Instruments Lab 2023
"""

from notochord import Notochord
from iipyper import MIDI, run, Stopwatch, cleanup

def main(
        player_channel=0, # MIDI channel numbered from 0
        noto_channel=1,
        player_inst=20, # General MIDI numbered from 1 (see Notochord.feed docstring)
        noto_inst=20,
        midi_in=None, # MIDI port for player input
        midi_out=None, # MIDI port for Notochord output
        checkpoint="notochord-latest.ckpt" # Notochord checkpoint
        ):
    midi = MIDI(midi_in, midi_out)

    if checkpoint is not None:
        noto = Notochord.from_checkpoint(checkpoint)
        noto.eval()
    else:
        noto = None

    note_map = {}
    stopwatch = Stopwatch()

    @midi.handle(type='program_change')
    def _(msg):
        """
        Program change events set instruments
        """
        nonlocal player_inst, noto_inst
        if msg.channel == player_channel:
            player_inst = msg.program
        if msg.channel == noto_channel:
            noto_inst = msg.program

    @midi.handle(type='control_change', control=0, channel=player_channel)
    def _(msg):
        """
        any CC0 message on player channel resets Notochord
        """
        noto.reset()

    @midi.handle(type=('note_on', 'note_off'), channel=player_channel)
    def _(msg):
        """
        MIDI NoteOn events from the player
        """
        if noto is None:
            print('Notochord model not loaded')
            return

        print('PLAYER:', msg)

        pitch = msg.note
        vel = msg.velocity

        # NoteOn
        if msg.type=='note_on' and vel > 0:
            # feed in the performed note
            noto.feed(player_inst, pitch, stopwatch.punch(), vel)
            # get the harmonizing note
            r = noto.query(
                next_inst=noto_inst, next_time=0, next_vel=vel,
                include_pitch=range(pitch+1, 128))
            print('NOTO:', r)
            # send it
            midi.note_on(
                note=r['pitch'], velocity=int(r['vel']), channel=noto_channel)
            # feed back
            noto.feed(r['inst'], r['pitch'], r['time'], r['vel'])
            # prepare for later NoteOff
            note_map[pitch] = r['pitch']
        # NoteOff
        else:
            try:
                noto_pitch = note_map.pop(pitch)
            except:
                print('harmonizing NoteOff not found')
                return
            # send harmonizing NoteOff
            midi.note_off(
                note=noto_pitch, velocity=vel, channel=noto_channel)
            # feed 
            noto.feed(player_inst, pitch, stopwatch.punch(), 0)
            noto.feed(noto_inst, noto_pitch, 0, 0)

    @cleanup
    def _():
        """end any remaining notes"""
        for pitch in note_map.values():
            midi.note_off(note=pitch, velocity=0, channel=noto_channel)

if __name__=='__main__':
    run(main)



from collections.abc import Mapping
import inspect
import urllib.request

import torch
from tqdm import tqdm

def argsort(arr):
    return sorted(range(len(arr)), key=lambda i: arr[i])

# from https://stackoverflow.com/a/53877507
class DownloadProgressBar(tqdm):
    def update_to(self, b=1, bsize=1, tsize=None):
        if tsize is not None:
            self.total = tsize
        self.update(b * bsize - self.n)
def download_url(url, output_path):
    with DownloadProgressBar(unit='B', unit_scale=True,
                             miniters=1, desc=url.split('/')[-1]) as t:
        urllib.request.urlretrieve(url, filename=output_path, reporthook=t.update_to)
##

def get_function_defaults(fn):
    """get dict of name:default for a function's arguments"""
    s = inspect.signature(fn)
    return {k:v.default for k,v in s.parameters.items()}

def get_class_defaults(cls):
    """get the default argument values of a class constructor"""
    d = get_function_defaults(getattr(cls, '__init__'))
    # ignore `self` argument, insist on default values
    try:
        d.pop('self')
    except KeyError:
        raise ValueError("""
            no `self` argument found in class __init__
        """)
    assert [v is not inspect._empty for v in d.values()], """
            get_class_defaults should be used on constructors with keyword arguments only.
        """
    return d

# mask for canonical order
def gen_perms(a):
    """yield all permutations of the given list"""
    if len(a)==1:
        yield a
    else:
        # for each position
        for i in range(len(a)):
            # for permuations of remaining positions
            for p in gen_perms(a[:i]+a[i+1:]):  
                yield a[i:i+1]+p 

def gen_masks(n, dtype=torch.float):
    """yield the autoregressive mask matrices of all permuations of n items"""
    for perm in gen_perms(list(range(n))):
        m = torch.zeros(n,n,dtype=dtype)
        for idx,i in enumerate(perm):
            for j in perm[:idx]:
                m[j,i] = 1
        yield perm, m


def deep_update(a, b):
    """
    in-place update a with contents of b, recursively for nested Mapping objects.
    """
    for k in b:
        if k in a and isinstance(a[k], Mapping) and isinstance(b[k], Mapping):
            deep_update(a[k], b[k])
        else:
            a[k] = b[k]


def arg_to_set(x):
    """convert None to empty set, iterable to set, or scalar to set with one item"""
    if x is None:
        return set()
    elif not hasattr(x, '__iter__'):
        return {x}
    else:
        return set(x)


"""
Authors:
  Victor Shepardson
  Jack Armitage
  Intelligent Instruments Lab 2022
"""

from notochord import Notochord
from iipyper import OSC, run
from iipyper.types import *
from time import time

def main(host="127.0.0.1", receive_port=9999, send_port=None, 
        checkpoint="notochord-latest.ckpt", prompt=None):
    osc = OSC(host, receive_port)

    initial_state = None

    if checkpoint is not None:
        predictor = Notochord.from_checkpoint(checkpoint)
        predictor.eval()
        predictor.reset()
        print('notochord created')
        if prompt is not None:
            initial_state, _ = predictor.prompt(prompt)
    else:
        predictor = None
 
    @osc.handle('/notochord/prompt', return_port=send_port)
    def _(address, midi_file:str, **kw):
        """
        load a MIDI file from a path and run it through the Notochord model.
        the hidden state after the prompt becomes the new reset state.
        """
        nonlocal initial_state
        print(f"{address} {midi_file} {kw}")
        initial_state, _ = predictor.prompt(midi_file, **kw)

    @osc.handle('/notochord/feed', return_port=send_port)
    def _(address, a:Splat[4], **kw):
        """
        Feed an event with given instrument, pitch, time and velocity
        For example: 
        /notochord/feed 1 60 0.2 99
            feed the model an event with instrument 1 (grand piano),
            MIDI number 60 (middle C), 200ms elapsed since last event,
            MIDI velocity 99
        """
        print(f"{address} {a} {kw}")
        predictor.feed(*a, **kw) 

    @osc.handle('/notochord/query', return_port=send_port)
    def _(address, **kw):
        """
        Arguments are given in key value pairs, mapping onto the Python API.
        For example:
        /notochord/query
            sample the next event with no constraints
        /notochord/query next_inst 1
            constrain next instrument to MIDI program 1 (grand piano)
        /notochord/query next_pitch 60
            constrain next pitch to MIDI number 60 (middle C)
        /notochord/query next_time 0.3
            constrain delta time from previous event to exactly 300ms
        /notochord/query next_vel 0
            constrain next velocity to 0 (noteOff)
        /notochord/query next_inst 129 next_vel 127
            constrain next instrument to 129 (standard drums),
            and next velocity to 127 (max)
        /notochord/query include_inst [ 1 2 3 ]
            constrain next instrument to be from within the given list 
            (client must support OSC arrays)
        /notochord/query pitch_temp 0.5 min_time 0.1
            query with sampling temperature for pitch reduced to 0.5,
            and constrain delta time since last event to at least 100ms

        for all options, see https://intelligent-instruments-lab.github.io/notochord/reference/notochord/model/#notochord.model.Notochord.query
        """
        print(f"{address} {kw}")
        r = predictor.query(**kw) 
        return (
            '/return/notochord/query', 
            *[x for pair in r.items() for x in pair])
    
    @osc.handle('/notochord/query_feed', return_port=send_port)
    def _(address, **kw):
        """
        /notochord/query_feed {feed arguments} {query arguments}
            query for an event and immediately feed it in one call
        """
        print(f"{address} {kw}")
        r = predictor.query_feed(**kw) 
        return (
            '/return/notochord/query', 
            *[x for pair in r.items() for x in pair])
    
    @osc.handle('/notochord/feed_query', return_port=send_port)
    def _(address, a:Splat[4], **kw):
        """
        /notochord/feed_query {feed arguments} {query arguments}
            feed an event, then query for the next in one call
        """
        print(f"{address} {a} {kw}")
        r = predictor.feed_query(*a, **kw) 
        return (
            '/return/notochord/query', 
            *[x for pair in r.items() for x in pair])

    @osc.handle('/notochord/reset', 
                return_port=send_port)
    def _(address, **kw):
        """
        /notochord/reset
            reset the model to the last prompted state, or else prime it with start-of-sequence token
        /notochord/reset "start" false
            reset the model without sending a start token
        """
        print(f"{address} {kw}")
        predictor.reset(state=initial_state, **kw) 

    @osc.handle('/notochord/load', 
                return_port=send_port)
    def _(address, path:str):
        """
        load your own pretrained notochord model
        /notochord/load "/path/to/my/notochord/model"
        """
        # `nonlocal` is needed to assign to closed-over name
        nonlocal predictor
        predictor = Notochord.from_checkpoint(path)
        predictor.eval()
        predictor.reset()
        print('notochord created')

    print('usage:')
    print(osc.get_docs())


if __name__=='__main__':
    run(main)



"""
Notochord + Language Model + Morse Code.

Authors:
  Victor Shepardson
  Intelligent Instruments Lab 2024
"""

import random
# from copy import deepcopy
import warnings
import logging

import rich
from rich.text import Text
import torch
from transformers import AutoModelForCausalLM, AutoTokenizer

from notochord import Notochord
from iipyper import MIDI, run, cleanup, repeat, _lock
import time

torch.set_num_threads(1)

def now():
    return time.time()

def main(
        noto_channel=1,
        noto_inst=20, # General MIDI numbered from 1 (see Notochord.feed docstring)
        midi_in=None, # MIDI port for player input
        midi_out=None, # MIDI port for Notochord output
        notochord="notochord-latest.ckpt", # Notochord checkpoint
        # lm="state-spaces/mamba-130m-hf",
        # lm="state-spaces/mamba-790m-hf",
        lm="state-spaces/mamba-1.4b-hf",
        verbose=0,
        prompt='morse code is ',
        dit_dur=0.05,
        send_pc=True,
        lm_temp=1,
        # noto_temp=1.5,
        noto_temp=1,
        noto_space=False, # does notochord contribute to prediction of ' ' char
        buffer_events=8,
        device='cpu',
        upper=True,
        ):
    midi = MIDI(midi_in, midi_out)

    print(f'loading notochord model {notochord}')
    noto = Notochord.from_checkpoint(notochord)
    noto.eval()

    # c_dot, c_dash, c_space = '•', '╺━╸', '┃'
    # c_dot, c_dash, c_space = '•', '╺╸', '┃'
    # c_dot, c_dash, c_space = '•', '⁃', '╱'
    c_dot, c_dash, c_space = '.', '╶╴', '╱'

    def do_send_pc(c, i):
        # warn_inst(i)
        # assuming fluidsynth -o synth.midi-bank-select=mma
        if noto.is_drum(i):
            midi.control_change(channel=c-1, control=0, value=1)
        else:
            midi.control_change(channel=c-1, control=32, value=0)
            midi.control_change(channel=c-1, control=0, value=0)
        if noto.is_anon(i):
            program = 0
        else:
            # convert to 0-index
            program = (i-1) % 128
        midi.program_change(channel=c-1, program=program)

    if send_pc:
        do_send_pc(noto_channel, noto_inst)

    print(f'loading language model {lm}')
    logging.disable()
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        tokenizer = AutoTokenizer.from_pretrained(lm)
        lm = AutoModelForCausalLM.from_pretrained(lm)
        lm.eval()
        lm.to(device)

    tokens = tokenizer.batch_decode(range(len(tokenizer)))
    alphabet = 'abcdefghijklmnopqrstuvwxyzðþöéæ"\'+-=/.,:;!?() '
    if upper:
        alphabet = alphabet.upper()
        prompt = prompt.upper()
        MorseNode.upper = True
    alphabet = set(alphabet)
    # indices of within-vocab tokens
    token_idx = [
        i for i,t in enumerate(tokens) 
        if all(c in alphabet for c in t)
        and '  ' not in t
        ]
    tokens = [tokens[i] for i in token_idx]
    # token_node_index maps tokens directly to tree nodes for speed
    token_root, token_node_index = TokenNode.make_tree(tokens)

    def set_token_probs(probs):
        for t,p in zip(tokens, probs):
            token_node_index[t].prob = p

    def lm_prompt(prompt):
        """feed a text to the language model and return probabilities for the next token and model state"""
        input_ids = tokenizer(prompt, return_tensors="pt")["input_ids"]
        input_ids = input_ids.to(lm.device)
        with torch.inference_mode():
            state = lm(input_ids, use_cache=True)
            probs = state.logits[0,-1,token_idx].softmax(0).cpu().tolist()
        return probs, state

    def lm_feed_query(token, state):
        """feed a token to the language model and return probabilities for the next token and model state"""
        with torch.inference_mode():
            input_ids = tokenizer(token, return_tensors='pt')['input_ids']
            input_ids = input_ids.to(lm.device)

            # print(input_ids, tokens[input_ids.item()])
            cache = None if state is None else state.cache_params
            # t = now()
            state = lm(input_ids, use_cache=True, cache_params=cache)
            probs = state.logits[0,-1,token_idx].softmax(0).cpu().tolist()
            # print('LM query ms:', int(1000*(now()-t)))

        return probs, state
    
    def set_morse_probs_from_token(morse_tree, token_tree):
        """set the probability of every character in morse to be the sum of probabilities of all currently possible tokens which begin with that character"""
        for node in morse_tree.nodes():
            node.prob = token_tree.children.get(
                node.char, MorseNode()).subtree_prob()
            if verbose > 3:
                print('    setting', node)

    def combine_probs(noto_dt_probs, lm_dt_probs):
        if len(noto_dt_probs)==3:
            noto_dt_probs = [noto_dt_probs[0], sum(noto_dt_probs[1:])]
        return [
            p1**(1/noto_temp) * p2**(1/lm_temp) 
            for p1, p2 in zip(noto_dt_probs, lm_dt_probs)]

    while prompt.startswith(' '):
        print('warning: prompt starts with space')
        prompt = prompt[1:]
    # TODO cache to disk so no delay when prompt is repeated
    # if prompt.endswith(' '):
    #     prompt = prompt[:-1]
    #     initial_space = True
    # else:
    #     initial_space = False
    # TODO get these in absence of prompt?? or they should be None?
    # initial_token_probs, initial_lm_state = lm_prompt(prompt)
    code = MorseNode.make_tree().make_code()
    # initial_morse = ' '.join(
    #     c_space if c==' ' else ''.join(c_dot if e else c_dash for e in code[c])
    #     for c in prompt
    #     )

    prompt_us = [0]
    prompt_token_ends = [False]
    prompt_tokens = iter(
        tokenizer.batch_decode(tokenizer(prompt)["input_ids"]))
    # prompt_u_tokens = [None]
    current_token = next(prompt_tokens)
    idx_in_token = 0
    for c in prompt:
        if c==' ':
            prompt_us[-1] = 7 # make it an interword gap
            # prompt_u_tokens[-1] = (current_token)
        else:
            for e in code[c]:
                prompt_us.append(1 if e else 3) # dit or dah
                prompt_us.append(1) # gap
                # prompt_u_tokens.extend([current_token]*2)
                prompt_token_ends.extend([False]*2)
            prompt_us[-1] = 3 # make it an intercharacter gap

        idx_in_token += 1
        if idx_in_token == len(current_token):
            current_token = next(prompt_tokens, None)
            idx_in_token = 0
            prompt_token_ends[-1] = True

        # print(f'{prompt_tokens=}')
        # print(f'{prompt_us=}')
        # print(f'{prompt_u_tokens=}')
        # print(f'{prompt_token_ends=}')

    class State:
        def __init__(self):
            self.reset()
        def reset(self):
            # self.next_event = None
            # self.next_event_time = None
            self.next_events = []
            self.next_event_times = []
            self.last_event_time = None
            self.noto_held_notes = set()
            self.synth_held_notes = set()

            self.token_probs = None#initial_token_probs
            self.lm_state = None#deepcopy(initial_lm_state)
            self.morse_tree = MorseNode.make_tree()

            self.morse = '' #initial_morse
            self.text = '' #prompt
            self.color_text = Text()
            self.current_token = ''
            self.is_gap = True
            # self.first_event = True
            
            self.token_tree = token_root
            # set_token_probs(self.token_probs)

            # self.token_tree.set_probs(token_idx, self.token_probs)
            # self.token_tree = TokenNode.make_tree(
                # tokens, self.token_probs)
            # self.token_tree = TokenNode.make_tree(
            #     tokens, self.token_probs, alphabet)
            set_morse_probs_from_token(self.morse_tree, self.token_tree)

    S = State()

    def release_all():
        """end any remaining notes"""
        for chan,pitch in S.synth_held_notes:
            midi.note_on(note=pitch, velocity=0, channel=chan-1)

    def reset():
        # TODO expose this
        noto.reset()
        S.reset()
        release_all()

    reset()

    def state_helper(event):
        k = (noto_channel, event['pitch'])
        if event['vel']>0:
            S.noto_held_notes.add(k)
        else:
            try:
                S.noto_held_notes.remove(k)
            except Exception:
                print(f'{k} not in noto_held_notes')
        S.is_gap = not S.is_gap

    def make_query():
        query = {}
        held_pitches = {p for _,p in S.noto_held_notes}
        if S.is_gap:
            # if determining the length of a gap,
            # that means we are predicting the next noteOff
            query['min_vel'] = 1
            query['include_pitch'] = set(range(20,100)) - held_pitches
            # query['next_pitch'] = 40
            # query['include_pitch'] = set(range(20,100)) - held_pitches
        else:
            # if determining the length of the tone,
            # that means we are predicting the next noteOff
            query['next_vel'] = 0
            # print(f'{held_pitches=}')
            if len(held_pitches)>1:
                query['include_pitch'] = held_pitches
            else:
                query['next_pitch'] = list(held_pitches)[0]
        # query['include_inst'] = [1, 8] 
        query['next_inst'] = noto_inst
        return query
    
    def process_event(event, u_possible, noto_dt_probs=None, token_end=None):
        if S.is_gap:
            if u_possible == [0]:
                if verbose > 0:
                    print('initial event')
            elif u_possible == [1]:
                u = 1
                if verbose > 0:
                    print('    character continues')
            else:
                c = S.morse_tree.char
                if verbose > 0:
                    print('  character complete:', c)
                event['text'] = c
                event['morse'] = ' '

                # put the new character into the current token
                S.token_tree = S.token_tree.traverse(c)
                # print(f'{token_tree.token} {"".join(token_tree.children)}')
                # possible end of token + LM query
                if token_end is None:
                    token_end = S.token_tree.sample_terminal(verbose=verbose)
                if token_end:
                    token = S.token_tree.token
                    if verbose > 0:
                        print('token complete:', token)
                    event['token'] = token
                    S.token_probs, S.lm_state = lm_feed_query(
                        token, S.lm_state)
                    S.token_tree = token_root
                    set_token_probs(S.token_probs)

                    if u_possible==[7]:
                        space = True
                    elif u_possible==[3]:
                        space = False
                    else:
                        # now sample nonspace vs space
                        noto_dt_probs = noto_dt_probs[1:]
                        space_prob = S.token_tree.children[' '].subtree_prob()
                        # don't needlessly traverse entire tree to sum probs
                        lm_dt_probs = [
                            sum(S.token_probs)-space_prob, space_prob]

                        if verbose > 1:
                            print(f'      LM prob [nonspace, space] {lm_dt_probs}')

                        if noto_space:
                            probs = combine_probs(noto_dt_probs, lm_dt_probs)
                        else:
                            probs = lm_dt_probs

                        space = random.random() < probs[1] / sum(probs)

                    if space:
                        u = 7
                        if verbose > 0:
                            print('space')
                        S.token_tree = S.token_tree.traverse(' ')
                        event['text'] += ' '
                        event['morse'] += c_space+' '
                    else:
                        u = 3
                        if verbose > 0:
                            print('word continues')

                else:
                    if verbose > 0:
                        print('  token continues')
                    u = 3

                # get next character probs from the token tree
                S.morse_tree = MorseNode.make_tree()
                set_morse_probs_from_token(S.morse_tree, S.token_tree)
                        
        else:
            assert len(u_possible)==1
            u = u_possible[0]
            S.morse_tree = S.morse_tree.traverse(u==1)
            event['morse'] = c_dot if u==1 else c_dash
            if verbose > 0:
                print('    dit' if u==1 else '    dah')

        return u

    def noto_prompt(u, token_end):
        """needs to know the duration of current event, plus whether a token is ending"""
        query = make_query()
        query['next_time'] = u*dit_dur
        # print(f'{u=} {token_end=} {S.is_gap=} {S.morse_tree=} {S.token_tree=}')

        event = noto.query_feed(**query)

        if u!=0:
            process_event(event, [u], token_end=token_end)

        state_helper(event)

        return event

    def noto_query():
        """query notochord + LM as needed for the next MIDI event
        """

        query = make_query()

        if S.is_gap:
            dts = [1,3,7]
        else:
            dts = [1,3]
        dts = torch.tensor([dts])*dit_dur

        with torch.inference_mode():
            # NOTE: would it be better to break apart the query?
            # condition on instrument, then sample velocity with constraint, 
            # then do the special sampling for time here,
            # then condition pitch on time
            # as it is i *think* it will do time last,
            # but should verify that.
            # should be possible to implement this using the new deep_query?
            # i.e. get the LM probs and pass them as weights for time Ranges
            # this would remove some flexibility in combine_probs though...
            # would also need to be able to pass in a 'case temperature' to deep query
            event = noto.query(return_params=True, **query)

            dt_breaks = torch.tensor([
                dit_dur*0.5, dit_dur*1.5, dit_dur*4.5])
            cdf = noto.time_dist.cdf(event['time_params'], dt_breaks).squeeze()
            noto_dt_probs = cdf[1:] - cdf[:-1]
            noto_dt_probs = noto_dt_probs.tolist()
            if S.is_gap:
                noto_dt_probs.append(1-cdf[-1].item())

            # time_sample = noto.time_dist(event['time_params'], dts)
            # noto_dt_probs = time_sample['log_prob'].exp().squeeze()
            # noto_dt_probs = [x.item() for x in noto_dt_probs]

        # print(f'{event["vel"]=}, {event["pitch"]=}')

        # get LM dt probs from tree
        if S.is_gap:
            if verbose > 1:
                print(f'      noto prob [1u, 3u, 7u] {noto_dt_probs}')
            # 1: char continues; 3,7: end of char
            lm_dt_probs = [S.morse_tree.children_prob(), S.morse_tree.prob]
            if verbose > 1:
                print(f'      LM prob [{S.morse_tree.children_chars()}, {S.morse_tree.char}] {lm_dt_probs}')
        else:
            if verbose > 1:
                print(f'      noto prob [1u, 3u] {noto_dt_probs}')
            # 1: dit, 3: dah
            dit = S.morse_tree.traverse(True)
            dah = S.morse_tree.traverse(False)
            lm_dt_probs = [dit.subtree_prob(), dah.subtree_prob()]
            if verbose > 1:
                print(f'      LM prob [dit, dah] [{dit.chars()}, {dah.chars()}] {lm_dt_probs}')

        # print(f'{"gap" if is_gap else "tone"} {lm_dt_probs=}')

        probs = combine_probs(noto_dt_probs, lm_dt_probs)
        if verbose > 1:
            print(f'      combined {probs=}')

        # here we sample 1 vs 3 or 1 vs 3,7
        is_1u = random.random() < probs[0] / (sum(probs) + 1e-10)

        u_possible = [1] if is_1u else ([3,7] if S.is_gap else [3])

        u = process_event(event, u_possible, noto_dt_probs=noto_dt_probs)

        event['time'] = u*dit_dur

        with torch.inference_mode():
            noto.feed(**event)
        
        state_helper(event)

        return event

    @midi.handle(type='program_change')
    def _(msg):
        """
        Program change events set instruments
        """
        nonlocal noto_inst
        if msg.channel == noto_channel:
            noto_inst = msg.program

    @midi.handle(type='control_change', control=0, channel=noto_channel-1)
    def _(msg):
        """
        any CC0 message resets Notochord
        """
        noto.reset()

    @midi.handle(type=('note_on', 'note_off'), channel=noto_channel-1)
    def _(msg):
        """
        MIDI NoteOn events
        """
        if noto is None:
            print('Notochord model not loaded')
            return
    
    # @repeat(1e-3, lock=True)

    color_generator = random.Random()
    def cg():
        return color_generator.randint(150,255)

    @repeat(lock=True)
    def _():
        """Loop, consuming scheduled events"""

        late = now() - S.next_event_times[0] if len(S.next_events) else -1
        if late >= 0:
            # if late > 10e-3:
                # print(f'{late=}')
            # send MIDI
            event = S.next_events.pop(0)
            pitch = int(event['pitch'])
            vel = int(event['vel'])
            midi.note_on(note=pitch, velocity=vel, channel=noto_channel-1)

            S.morse += event.get('morse', '')
            char = event.get('text', '')
            # S.text += event.get('text', '')

            if 'token' in event:
                tok = event['token']
                color_generator.seed(tok)
                S.color_text.append(event['token'], style=f'rgb({cg()},{cg()},{cg()})')
                S.current_token = ' ' if char.endswith(' ') else ''
            else:
                S.current_token += char

            rich.print(
                Text('\n'+S.morse+'\n')+S.color_text+Text(S.current_token), 
                sep='', end='')

            # print(
                # '\n'+S.morse+'\n'+S.text, 
                # flush=True, end='')

            k = (noto_channel, pitch)
            if vel:
                S.synth_held_notes.add(k)
            else:
                try:
                    S.synth_held_notes.remove(k)
                except Exception:
                    print(f'{k} not in synth_held_notes')

            S.last_event_time = S.next_event_times.pop(0)

        dt = S.next_event_times[0] - now() if len(S.next_event_times) else 10e-3
        return max(0, min(10e-3, dt))

    @repeat(10e-3, lock=False)
    def _():
        """Loop, predicting next events"""          
        if len(S.next_events) < buffer_events:
            final_time = (
                S.next_event_times[-1] if len(S.next_event_times) 
                else S.last_event_time)
            
            # t = now()
            # if final_time is not None:
                # print('buffered time ms:', int(1000*(final_time-t)))
            if len(prompt_us):
                u = prompt_us.pop(0)
                # token = prompt_u_tokens.pop(0)
                token_end = prompt_token_ends.pop(0)
                event = noto_prompt(u, token_end) 
            else:
                event = noto_query()
            # print('query time ms:', int(1000*(now()-t)))

            with _lock:
                S.next_events.append(event)

                if final_time is None:
                    S.next_event_times.append(now() + 2)
                else:
                    S.next_event_times.append(final_time + float(event['time']))

        # print('buffered events:', len(S.next_events))

        # return max(0, min(10e-3, S.next_event_times[0] - now()))
        # return max(0, min(10e-3, S.next_event_time - now()))


    @cleanup
    def _():
        release_all()

class MorseNode:
    upper:bool = False
    def __init__(self, 
            char:str=None, 
            prob:float=0, 
            dit:'MorseNode'=None, 
            dah:'MorseNode'=None
        ):
        if char is not None and MorseNode.upper:
            char = char.upper()
        self.char = char
        self.prob = prob
        self.dit = dit
        self.dah = dah

    def subtree_prob(self):
        prob = self.prob
        for d in (self.dit, self.dah):
            if d is not None: 
                prob += d.subtree_prob()
        return prob
    
    def children_prob(self):
        return self.subtree_prob() - self.prob
    
    def is_leaf(self):
        return self.dit is None and self.dah is None
    
    def nodes(self):
        nodes = [self]
        for d in (self.dit, self.dah):
            if d is not None:
                nodes.extend(d.nodes())
        return nodes
    
    def chars(self):
        return ''.join(
            MorseNode.char for MorseNode in self.nodes() if MorseNode.char is not None)
    
    def children_chars(self):
        chars = ''
        for d in (self.dit, self.dah):
            if d is not None:
                chars += d.chars()
        return chars
    
    def traverse(self, *seq):
        if not len(seq):
            return self
        b, *seq = seq
        d = self.dit if b else self.dah
        # print(b, seq, d)
        return MorseNode() if d is None else d.traverse(*seq)
        
    def __repr__(self):
        s = f'{self.char} {self.prob}'
        for d in (self.dit, self.dah):
            if d is not None: 
                s += f' {d.char}'
        return s
    
    def make_code(self, path=None):
        """return a mapping from characters to morse"""
        code = {}
        if path is None: path = tuple()
        if self.char is not None:
            code[self.char] = path
        for d,b in ((self.dit, True), (self.dah, False)):
            if d is not None:
                code.update(d.make_code(path+(b,)))
        return code

    @classmethod
    def make_tree(cls):
        return MorseNode(
            dit=MorseNode('e', 12.7,
                dit=MorseNode('i', 7.0,
                    dit=MorseNode('s', 6.3,
                        dit=MorseNode('h', 6.1,
                            dit=MorseNode('5'),
                            dah=MorseNode('4'),
                        ),
                        dah=MorseNode('v', 1.0,
                            # dit=MorseNode('ŝ'),
                            dah=MorseNode('3'),
                        ),
                    ),
                    dah=MorseNode('u', 2.8,
                        dit=MorseNode('f', 2.2,
                            dit=MorseNode('é'),
                            dah=MorseNode('4'),
                        ),
                        dah=MorseNode(#'ü',
                            dit=MorseNode('ð',
                                dit=MorseNode('?')
                            ),
                            dah=MorseNode('2'),
                        ),
                    ),
                ),
                dah=MorseNode('a', 8.2,
                    dit=MorseNode('r', 6.0,
                        dit=MorseNode('l', 4.0,
                            # dit=MorseNode(),
                            dah=MorseNode(#'è',
                                dit=MorseNode('"')
                            ),
                        ),
                        dah=MorseNode('æ',
                            dit=MorseNode('+',
                                dah=MorseNode('.')
                            )
                        ),
                    ),
                    dah=MorseNode('w', 2.4,
                        dit=MorseNode('p', 1.9,
                            dit=MorseNode('þ'),
                            dah=MorseNode(#'à',
                                dit=MorseNode('@'),
                            ),
                        ),
                        dah=MorseNode('j', 0.2,
                            # dit=MorseNode('ĵ'),
                            dah=MorseNode('1',
                                dit=MorseNode("'")
                            ),
                        ),
                    ),
                ),
            ),
            dah=MorseNode('t', 9.1,
                dit=MorseNode('n', 6.7,
                    dit=MorseNode('d', 4.3,
                        dit=MorseNode('b', 1.5,
                            dit=MorseNode('6',
                                dah=MorseNode('-')
                            ),
                            dah=MorseNode('='),
                        ),
                        dah=MorseNode('x', 0.2,
                            dit=MorseNode('/')
                        ),
                    ),
                    dah=MorseNode('k', 0.8,
                        dit=MorseNode('c', 2.8,
                            # dit=MorseNode('ç'),
                            dah=MorseNode(#empty MorseNode
                                dit=MorseNode(';'),
                                dah=MorseNode('!')
                            ),
                        ),
                        dah=MorseNode('y', 2.0,
                            dit=MorseNode('(',
                                dah=MorseNode(')')
                            )
                        ),
                    ),
                ),
                dah=MorseNode('m', 2.4,
                    dit=MorseNode('g', 2.0,
                        dit=MorseNode('z', 0.1,
                            dit=MorseNode('7'),
                            dah=MorseNode(
                                dah=MorseNode(',')
                            ),
                        ),
                        dah=MorseNode('q', 0.1),
                    ),
                    dah=MorseNode('o', 7.5,
                        dit=MorseNode('ö',
                            dit=MorseNode('8',
                                dit=MorseNode(':')
                            ),         
                        ),
                        dah=MorseNode(
                            dit=MorseNode('9'),
                            dah=MorseNode('0')
                        ),
                    ),
                ),
            )
        )

class TokenNode:
    def __init__(self, 
            token:str=None, 
            prob:float=0, 
            children=None,
        ):
        self.token = token
        self.prob = prob
        if children is None:
            children = {}
        self.children = children
            
    def subtree_prob(self):
        prob = self.prob
        for node in self.children.values():
            prob += node.subtree_prob()
        return prob
    
    def children_prob(self):
        return self.subtree_prob() - self.prob
    
    def is_leaf(self):
        return not len(self.children)
    
    def traverse(self, s):
        if not len(s):
            return self
        c, *s = s
        node = self.children.get(c)
        return TokenNode() if node is None else node.traverse(s)
    
    def nodes(self):
        yield self
        for child in self.children.values():
            for node in child.nodes():
                yield node
    
    def sample_terminal(self, verbose=0):
        if verbose>2:
            # print(self)
            print(f'----terminal prob: {self.prob / self.subtree_prob()}')
        if self.prob > 0:
            if random.random() < self.prob / self.subtree_prob():
                return True
        return False
        
    def __repr__(self):
        s = f'{self.token} {self.prob} {"".join(self.children)}'
        return s

    @classmethod
    def make_tree(cls, tokens):
        root = TokenNode()
        index = {}
        for token in tokens:
            tree = root
            for c in token:
                if c not in tree.children:
                    tree.children[c] = TokenNode()
                tree = tree.children[c]
            tree.token = token
            index[token] = tree
        return root, index

if __name__=='__main__':
    run(main)
